import mock
import colander
from cornice.validators import colander_validator
from pyramid import exceptions
from pyramid import testing

from kinto.core import authorization, DEFAULT_SETTINGS
from kinto.core.resource import ViewSet, ShareableViewSet, register_resource
from kinto.core.resource.viewset import PartialSchema, StrictSchema
from kinto.core.testing import unittest


class FakeViewSet(ViewSet):
    """Fake viewset class used for tests."""
    collection_path = "/{resource_name}"
    record_path = "/{resource_name}/{{id}}"

    collection_methods = ('GET',)
    record_methods = ('PUT',)

    def __init__(self):
        self.collection_arguments = self.arguments
        self.record_arguments = self.arguments
        self.update = mock.MagicMock()
        self.responses = mock.MagicMock()

    def arguments(self, resource, method):
        # By default, returns an empty dict of arguments.
        return {}


class FakeResource:
    """Fake resource class used for tests"""
    name = "fake"

    def __init__(self):
        # Create fake views which are in fact mock sentinels.
        # {type}_{method} will map to the sentinel with the same name.
        for endpoint_type in ('collection', 'record'):
            for method in ('get', 'put', 'patch', 'delete'):
                if endpoint_type == 'record':
                    view_name = method
                else:
                    view_name = '_'.join((endpoint_type, method))
                setattr(self, view_name, getattr(mock.sentinel, view_name))


class ViewSetTest(unittest.TestCase):

    def test_arguments_are_merged_on_initialization(self):
        viewset = ViewSet(collection_path=mock.sentinel.collection_path)
        self.assertEquals(viewset.collection_path,
                          mock.sentinel.collection_path)

    def test_default_arguments_are_copied_before_being_returned(self):
        original_arguments = {}
        viewset = ViewSet(
            collection_get_arguments=original_arguments)
        viewset.responses = mock.MagicMock()
        arguments = viewset.collection_arguments(mock.MagicMock(), 'GET')
        self.assertEquals(original_arguments, {})
        self.assertNotEquals(original_arguments, arguments)

    def test_permission_private_is_set_by_default(self):
        viewset = ViewSet()
        viewset.responses = mock.MagicMock()
        args = viewset.collection_arguments(mock.MagicMock(), 'GET')
        self.assertEquals(args['permission'], 'private')

    def test_schema_is_added_when_method_matches(self):
        viewset = ViewSet()
        viewset.responses = mock.MagicMock()
        resource = mock.MagicMock()
        arguments = viewset.collection_arguments(resource, 'GET')
        self.assertIn('schema', arguments)

    def test_schema_is_added_when_uppercase_method_matches(self):
        viewset = ViewSet(
            collection_methods=('GET', 'DELETE')
        )
        viewset.responses = mock.MagicMock()
        arguments = viewset.collection_arguments(mock.MagicMock(), 'get')
        self.assertIn('schema', arguments)

    @mock.patch('kinto.core.resource.viewset.RequestSchema')
    def test_a_default_schema_is_added_when_method_doesnt_match(self, mocked):
        viewset = ViewSet()
        resource = mock.MagicMock()
        viewset.responses = mock.MagicMock()
        mocked.Mapping.return_value = mock.sentinel.default_schema

        arguments = viewset.collection_arguments(resource, 'GET')
        self.assertNotEqual(arguments['schema'], resource.schema)

        mocked.assert_called_with()

    @mock.patch('kinto.core.resource.viewset.RequestSchema')
    def test_class_parameters_are_used_for_collection_arguments(self, mocked):
        default_arguments = {
            'cors_headers': mock.sentinel.cors_headers,
        }

        default_get_arguments = {
            'accept': mock.sentinel.accept,
        }

        default_collection_arguments = {
            'cors_origins': mock.sentinel.cors_origins,
        }

        collection_get_arguments = {
            'error_handler': mock.sentinel.error_handler
        }

        viewset = ViewSet(
            default_arguments=default_arguments,
            default_get_arguments=default_get_arguments,
            default_collection_arguments=default_collection_arguments,
            collection_get_arguments=collection_get_arguments
        )

        viewset.responses = mock.MagicMock()
        arguments = viewset.collection_arguments(mock.MagicMock(), 'get')

        self.assertDictEqual(
            arguments,
            {
                'schema': mocked().bind(),
                'accept': mock.sentinel.accept,
                'cors_headers': mock.sentinel.cors_headers,
                'cors_origins': mock.sentinel.cors_origins,
                'error_handler': mock.sentinel.error_handler,
                'validators': [colander_validator],
                'response_schemas': viewset.responses.get_and_bind()
            }
        )

    @mock.patch('kinto.core.resource.viewset.RequestSchema')
    def test_default_arguments_are_used_for_record_arguments(self, mocked):
        default_arguments = {
            'cors_headers': mock.sentinel.cors_headers,
        }

        default_get_arguments = {
            'accept': mock.sentinel.accept,
        }

        default_record_arguments = {
            'cors_origins': mock.sentinel.record_cors_origins,
        }

        record_get_arguments = {
            'error_handler': mock.sentinel.error_handler
        }

        viewset = ViewSet(
            default_arguments=default_arguments,
            default_get_arguments=default_get_arguments,
            default_record_arguments=default_record_arguments,
            record_get_arguments=record_get_arguments
        )

        viewset.responses = mock.MagicMock()
        arguments = viewset.record_arguments(mock.MagicMock(), 'get')

        self.assertDictEqual(
            arguments,
            {
                'schema': mocked().bind(),
                'accept': mock.sentinel.accept,
                'cors_headers': mock.sentinel.cors_headers,
                'cors_origins': mock.sentinel.record_cors_origins,
                'error_handler': mock.sentinel.error_handler,
                'validators': [colander_validator],
                'response_schemas': viewset.responses.get_and_bind()
            }
        )

    @mock.patch('kinto.core.resource.viewset.RequestSchema')
    def test_class_parameters_overwrite_each_others(self, mocked):
        # Some class parameters should overwrite each others.
        # The more specifics should prevail over the more generics.
        # Items annoted with a "<<" are the one that should prevail.
        default_arguments = {
            'cors_origins': mock.sentinel.default_cors_origins,
            'error_handler': mock.sentinel.default_error_handler,
            'cors_headers': mock.sentinel.default_cors_headers,  # <<
        }
        default_record_arguments = {
            'cors_origins': mock.sentinel.default_record_cors_origin,
            'error_handler': mock.sentinel.default_record_error_handler,  # <<
        }

        record_get_arguments = {
            'cors_origins': mock.sentinel.record_get_cors_origin,  # <<
        }

        viewset = ViewSet(
            default_arguments=default_arguments,
            default_record_arguments=default_record_arguments,
            record_get_arguments=record_get_arguments,
        )

        viewset.responses = mock.MagicMock()
        arguments = viewset.record_arguments(mock.MagicMock(), 'get')

        self.assertDictEqual(
            arguments,
            {
                'schema': mocked().bind(),
                'cors_headers': mock.sentinel.default_cors_headers,
                'error_handler': mock.sentinel.default_record_error_handler,
                'cors_origins': mock.sentinel.record_get_cors_origin,
                'validators': [colander_validator],
                'response_schemas': viewset.responses.get_and_bind()
            }
        )

    @mock.patch('kinto.core.resource.viewset.RequestSchema')
    def test_service_arguments_arent_inherited_by_record_arguments(self, mocked):
        service_arguments = {
            'description': 'The little book of calm',
        }

        default_arguments = {
            'cors_headers': mock.sentinel.cors_headers,
        }
        viewset = ViewSet(
            default_arguments=default_arguments,
            service_arguments=service_arguments,
            default_record_arguments={},
            record_get_arguments={}
        )

        viewset.responses = mock.MagicMock()
        arguments = viewset.record_arguments(mock.MagicMock(), 'get')

        self.assertDictEqual(
            arguments,
            {
                'schema': mocked().bind(),
                'cors_headers': mock.sentinel.cors_headers,
                'validators': [colander_validator],
                'response_schemas': viewset.responses.get_and_bind()
            }
        )

    def test_get_service_name_returns_the_viewset_name_if_defined(self):
        viewset = ViewSet(name='fakename')
        self.assertEquals(
            viewset.get_service_name('record', mock.MagicMock),
            'fakename-record')

    def test_get_service_name_returns_resource_att_if_not_callable(self):
        viewset = ViewSet()
        resource = mock.MagicMock()
        resource.name = 'fakename'
        self.assertEquals(
            viewset.get_service_name('record', resource),
            'fakename-record')

    def test_get_service_name_doesnt_use_callable_as_a_name(self):
        viewset = ViewSet()
        resource = mock.MagicMock()
        resource.name = lambda x: 'should not be called'
        resource.__name__ = "FakeName"
        self.assertEquals(
            viewset.get_service_name('record', resource),
            'fakename-record')

    def test_get_service_arguments_has_no_factory_by_default(self):
        viewset = ViewSet()
        service_arguments = viewset.get_service_arguments()
        self.assertNotIn('factory', service_arguments)

    def test_is_endpoint_enabled_returns_true_if_unknown(self):
        viewset = ViewSet()
        config = {}
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'get',
                                                 config)
        self.assertTrue(is_enabled)

    def test_is_endpoint_enabled_returns_false_if_disabled(self):
        viewset = ViewSet()
        config = {
            'record_fake_get_enabled': False
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'get',
                                                 config)
        self.assertFalse(is_enabled)

    def test_is_endpoint_enabled_returns_true_if_enabled(self):
        viewset = ViewSet()
        config = {
            'record_fake_get_enabled': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'get',
                                                 config)
        self.assertTrue(is_enabled)

    def test_is_endpoint_enabled_returns_false_for_put_if_readonly(self):
        viewset = ViewSet()
        config = {
            'readonly': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'put',
                                                 config)
        self.assertFalse(is_enabled)

    def test_is_endpoint_enabled_returns_false_for_post_if_readonly(self):
        viewset = ViewSet()
        config = {
            'readonly': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'post',
                                                 config)
        self.assertFalse(is_enabled)

    def test_is_endpoint_enabled_returns_false_for_patch_if_readonly(self):
        viewset = ViewSet()
        config = {
            'readonly': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'patch',
                                                 config)
        self.assertFalse(is_enabled)

    def test_is_endpoint_enabled_returns_false_for_delete_if_readonly(self):
        viewset = ViewSet()
        config = {
            'readonly': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'delete',
                                                 config)
        self.assertFalse(is_enabled)

    def test_is_endpoint_enabled_returns_true_for_get_if_readonly(self):
        viewset = ViewSet()
        config = {
            'readonly': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'get',
                                                 config)
        self.assertTrue(is_enabled)

    def test_is_endpoint_enabled_returns_true_for_options_if_readonly(self):
        viewset = ViewSet()
        config = {
            'readonly': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'options',
                                                 config)
        self.assertTrue(is_enabled)

    def test_is_endpoint_enabled_returns_true_for_head_if_readonly(self):
        viewset = ViewSet()
        config = {
            'readonly': True
        }
        is_enabled = viewset.is_endpoint_enabled('record', 'fake', 'head',
                                                 config)
        self.assertTrue(is_enabled)


class TestViewsetBindedSchemas(unittest.TestCase):

    def setUp(self):
        self.viewset = ViewSet()
        self.viewset.responses = mock.MagicMock()
        self.resource = mock.MagicMock()

    def test_request_schemas_have_header_and_querystring(self):
        self.viewset = ViewSet(
            default_get_arguments={},
            default_record_arguments={},
            record_get_arguments={}
        )
        arguments = self.viewset.record_arguments(self.resource, 'GET')
        schema = arguments['schema']
        self.assertIn('querystring', schema)
        self.assertIn('header', schema)

    def test_payload_request_schemas_have_a_body(self):
        arguments = self.viewset.record_arguments(self.resource, 'PUT')
        schema = arguments['schema']
        self.assertIn('body', schema)

    def test_collection_deserialize_sort(self):
        arguments = self.viewset.collection_arguments(self.resource, 'DELETE')
        schema = arguments['schema']
        value = {'querystring': {'_sort': 'foo,-bar'}}
        deserialized = schema.deserialize(value)
        expected = {'querystring': {'_sort': ['foo', '-bar']}}
        self.assertEquals(deserialized, expected)

    def test_get_collection_deserialize_fields(self):
        arguments = self.viewset.collection_arguments(self.resource, 'GET')
        schema = arguments['schema']
        value = {'querystring': {'_fields': 'foo,bar'}}
        deserialized = schema.deserialize(value)
        expected = {'querystring': {'_fields': ['foo', 'bar']}}
        self.assertEquals(deserialized, expected)

    def test_get_record_deserialize_fields(self):
        arguments = self.viewset.record_arguments(self.resource, 'GET')
        schema = arguments['schema']
        value = {'querystring': {'_fields': 'foo,bar'}}
        deserialized = schema.deserialize(value)
        expected = {'querystring': {'_fields': ['foo', 'bar']}}
        self.assertEquals(deserialized, expected)

    def test_patch_record_validate_response_behavior(self):
        arguments = self.viewset.collection_arguments(self.resource, 'PATCH')
        schema = arguments['schema']
        invalid = {'header': {'Response-Behavior': 'impolite'}}
        self.assertRaises(colander.Invalid, schema.deserialize, invalid)


class TestViewsetSchemasTest(unittest.TestCase):

    def test_partial_schema_ignores_unknown(self):
        schema = PartialSchema()
        result = schema.deserialize({'foo': 'bar'})
        self.assertEquals(result, {})

    def test_strict_schema_raise_on_unknown(self):
        schema = StrictSchema()
        self.assertRaises(colander.Invalid, schema.deserialize, {'foo': 'bar'})


class ShareableViewSetTest(unittest.TestCase):

    def test_permission_dynamic_is_set_by_default(self):
        viewset = ShareableViewSet()
        viewset.responses = mock.MagicMock()
        resource = mock.MagicMock()
        args = viewset.collection_arguments(resource, 'GET')
        self.assertEquals(args['permission'], 'dynamic')

    def test_get_service_arguments_has_default_factory(self):
        viewset = ShareableViewSet()
        args = viewset.get_service_arguments()
        self.assertEqual(args['factory'], authorization.RouteFactory)

    def test_mapping_is_deprecated(self):
        viewset = ShareableViewSet()
        viewset.responses = mock.MagicMock()
        resource = mock.MagicMock()
        resource.mapping = mock.MagicMock()
        with mock.patch('kinto.core.resource.viewset.warnings') as mocked:
            viewset.collection_arguments(resource, 'GET')
            msg = "Resource `mapping` is deprecated, use `schema`"
            mocked.warn.assert_called_with(msg, DeprecationWarning)


class RegisterTest(unittest.TestCase):

    def setUp(self):
        self.resource = FakeResource
        self.viewset = FakeViewSet()

    @mock.patch('kinto.core.resource.Service')
    def test_register_fails_if_no_storage_backend_is_configured(self, *args):
        venusian_callback = register_resource(
            self.resource, viewset=self.viewset)
        context = mock.MagicMock()
        config = testing.setUp(settings=DEFAULT_SETTINGS)
        context.config.with_package.return_value = config
        try:
            venusian_callback(context, None, None)
        except exceptions.ConfigurationError as e:
            error = e
        self.assertIn('storage backend is missing', str(error))

    @mock.patch('kinto.core.resource.Service')
    def test_viewset_is_updated_if_provided(self, service_class):
        additional_params = {'foo': 'bar'}
        register_resource(self.resource, viewset=self.viewset,
                          **additional_params)
        self.viewset.update.assert_called_with(**additional_params)

    def test_resource_default_viewset_is_used_if_not_provided(self):
        resource = FakeResource
        resource.default_viewset = mock.Mock()
        additional_params = {'foo': 'bar'}
        register_resource(resource, **additional_params)
        resource.default_viewset.assert_called_with(**additional_params)

    @mock.patch('kinto.core.resource.Service')
    def test_collection_views_are_registered_in_cornice(self, service_class):
        venusian_callback = register_resource(
            self.resource, viewset=self.viewset)

        config = mock.MagicMock()
        config.registry.settings = DEFAULT_SETTINGS

        context = mock.MagicMock()
        context.config.with_package.return_value = config
        venusian_callback(context, None, None)

        service_class.assert_any_call('fake-collection', '/fake', depth=1,
                                      **self.viewset.get_service_arguments())
        service_class().add_view.assert_any_call(
            'GET', 'collection_get', klass=self.resource)

    @mock.patch('kinto.core.resource.Service')
    def test_record_views_are_registered_in_cornice(self, service_class):
        venusian_callback = register_resource(
            self.resource, viewset=self.viewset)

        config = mock.MagicMock()
        config.registry.settings = DEFAULT_SETTINGS

        context = mock.MagicMock()
        context.config.with_package.return_value = config
        venusian_callback(context, None, None)

        service_class.assert_any_call('fake-record', '/fake/{id}', depth=1,
                                      **self.viewset.get_service_arguments())
        service_class().add_view.assert_any_call(
            'PUT', 'put', klass=self.resource)

    @mock.patch('kinto.core.resource.Service')
    def test_collection_methods_are_skipped_if_not_enabled(self, service_cls):
        venusian_callback = register_resource(
            self.resource, viewset=self.viewset)

        context = mock.MagicMock()
        context.registry.settings = {
            'record_fake_put_enabled': False
        }
        context.config.with_package.return_value = context
        venusian_callback(context, None, None)

        # Only the collection views should have been added.
        # 3 calls: two registering the service classes,
        # one for the collection_get
        self.assertEquals(len(service_cls.mock_calls), 3)
        service_cls.assert_any_call('fake-collection', '/fake', depth=1,
                                    **self.viewset.get_service_arguments())
        service_cls().add_view.assert_any_call(
            'GET', 'collection_get', klass=self.resource)

    @mock.patch('kinto.core.resource.Service')
    def test_record_methods_are_skipped_if_not_enabled(self, service_class):
        venusian_callback = register_resource(
            self.resource, viewset=self.viewset)

        context = mock.MagicMock()
        context.config.with_package.return_value = context
        context.registry.settings = {
            'collection_fake_get_enabled': False
        }
        venusian_callback(context, None, None)

        # Only the collection views should have been added.
        # 3 calls: two registering the service classes,
        # one for the collection_get
        self.assertEquals(len(service_class.mock_calls), 3)
        service_class.assert_any_call('fake-record', '/fake/{id}', depth=1,
                                      **self.viewset.get_service_arguments())
        service_class().add_view.assert_any_call(
            'PUT', 'put', klass=self.resource)

    @mock.patch('kinto.core.resource.Service')
    def test_endpoint_is_skipped_if_record_path_is_none(self, service_cls):
        self.viewset.record_path = None

        venusian_callback = register_resource(self.resource,
                                              viewset=self.viewset)
        config = mock.MagicMock()
        config.registry.settings = DEFAULT_SETTINGS
        context = mock.MagicMock()
        context.config.with_package.return_value = config
        venusian_callback(context, None, None)

        paths = [call[1][1] for call in service_cls.mock_calls]
        self.assertIn('/fake', paths)
        self.assertNotIn('/fake/{id}', paths)



import datetime

import mock

from kinto.core.errors import ERRORS
from kinto.core.utils import json
from kinto.core.testing import unittest

from .support import BaseWebTest


class DeprecationTest(BaseWebTest, unittest.TestCase):

    def test_do_not_return_alert_if_no_eos(self):
        response = self.app.get('/')
        self.assertNotIn('Alert', response.headers)

    def test_returns_alert_if_eos_in_the_future(self):
        # Set an end of service date in the future.
        tomorrow = datetime.datetime.now() + datetime.timedelta(days=1)
        with mock.patch.dict(self.app.app.registry.settings, [
                             ('eos', tomorrow.isoformat()),
                             ('eos_url', 'http://eos-url'),
                             ('eos_message',
                              'This service will soon be decommissioned')]):
            response = self.app.get('/')

            # Requests should work as usual and contain an
            # Alert header, with the service end of life information.
            self.assertIn('Alert', response.headers)
            self.assertDictEqual(json.loads(response.headers['Alert']), {
                'code': 'soft-eol',
                'url': 'http://eos-url',
                'message': 'This service will soon be decommissioned'
            })

    def test_returns_410_if_eos_in_the_past(self):
        # Set an end of service date in the past.
        yesterday = datetime.datetime.now() - datetime.timedelta(days=1)
        with mock.patch.dict(self.app.app.registry.settings, [
                             ('eos', yesterday.isoformat()),
                             ('eos_url', 'http://eos-url'),
                             ('eos_message',
                              'This service had been decommissioned')]):
            response = self.app.get('/', status=410)
            self.assertIn('Alert', response.headers)
            self.assertDictEqual(json.loads(response.headers['Alert']), {
                'code': 'hard-eol',
                'url': 'http://eos-url',
                'message': 'This service had been decommissioned'
            })
            self.assertDictEqual(response.json, {
                "errno": ERRORS.SERVICE_DEPRECATED.value,
                "message": "The service you are trying to connect no longer "
                           "exists at this location.",
                "code": 410,
                "error": "Gone"
            })



import random

from pyramid import httpexceptions

from kinto.core.errors import ERRORS

from . import BaseTest


class SortingTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()

        indices = list(range(20))
        random.shuffle(indices)

        for i in indices:
            record = {
                'title': 'MoFo #{0:02}'.format(i),
                'status': i % 4,
                'unread': (i % 2 == 0)
            }
            self.model.create_record(record)

    def test_sort_works_with_empty_list(self):
        self.resource.model.parent_id = 'alice'
        self.resource.request.validated['querystring'] = {'_sort': ['unread']}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 0)

    def test_sort_on_unknown_attribute_raises_error(self):
        self.patch_known_field.stop()
        self.resource.request.validated['querystring'] = {'_sort': ['foo']}
        self.assertRaises(httpexceptions.HTTPBadRequest,
                          self.resource.collection_get)

    def test_filter_errors_are_json_formatted(self):
        self.patch_known_field.stop()
        self.resource.request.validated['querystring'] = {'_sort': ['foo']}
        try:
            self.resource.collection_get()
        except httpexceptions.HTTPBadRequest as e:
            error = e
        self.assertEqual(error.json, {
            'errno': ERRORS.INVALID_PARAMETERS.value,
            'message': "querystring: Unknown sort field 'foo'",
            'code': 400,
            'error': 'Invalid parameters',
            'details': [{'description': "Unknown sort field 'foo'",
                         'location': 'querystring',
                         'name': None}]})

    def test_sort_on_last_modified_is_supported(self):
        self.patch_known_field.stop()
        self.resource.request.validated['querystring'] = {'_sort': ['-last_modified']}
        result = self.resource.collection_get()
        tstamp = self.model.timestamp()
        self.assertEqual(result['data'][0]['last_modified'], tstamp)

    def test_single_basic_sort_by_attribute(self):
        self.resource.request.validated['querystring'] = {'_sort': ['title']}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 20)
        self.assertEqual(result['data'][0]['title'], 'MoFo #00')
        self.assertEqual(result['data'][-1]['title'], 'MoFo #19')

    def test_single_basic_sort_by_attribute_reversed(self):
        self.resource.request.validated['querystring'] = {'_sort': ['-title']}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 20)
        self.assertEqual(result['data'][0]['title'], 'MoFo #19')
        self.assertEqual(result['data'][-1]['title'], 'MoFo #00')

    def test_multiple_sort(self):
        self.resource.request.validated['querystring'] = {'_sort': ['status', 'title']}
        result = self.resource.collection_get()
        self.assertEqual(result['data'][0]['status'], 0)
        self.assertEqual(result['data'][0]['title'], 'MoFo #00')
        self.assertEqual(result['data'][1]['status'], 0)
        self.assertEqual(result['data'][1]['title'], 'MoFo #04')
        self.assertEqual(result['data'][-2]['status'], 3)
        self.assertEqual(result['data'][-2]['title'], 'MoFo #15')
        self.assertEqual(result['data'][-1]['status'], 3)
        self.assertEqual(result['data'][-1]['title'], 'MoFo #19')

    def test_multiple_sort_with_order(self):
        self.resource.request.validated['querystring'] = {'_sort': ['status', '-title']}
        result = self.resource.collection_get()
        self.assertEqual(result['data'][0]['status'], 0)
        self.assertEqual(result['data'][0]['title'], 'MoFo #16')
        self.assertEqual(result['data'][1]['status'], 0)
        self.assertEqual(result['data'][1]['title'], 'MoFo #12')
        self.assertEqual(result['data'][-2]['status'], 3)
        self.assertEqual(result['data'][-2]['title'], 'MoFo #07')
        self.assertEqual(result['data'][-1]['status'], 3)
        self.assertEqual(result['data'][-1]['title'], 'MoFo #03')

    def test_boolean_sort_brings_true_last(self):
        self.resource.request.validated['querystring'] = {'_sort': ['unread']}
        result = self.resource.collection_get()
        self.assertEqual(result['data'][0]['unread'], False)
        self.assertEqual(result['data'][-1]['unread'], True)

    def test_default_sort_is_last_modified_desc_by_default(self):
        self.resource.request.validated['querystring'] = {'_sort': ['']}
        result = self.resource.collection_get()
        tstamp = self.model.timestamp()
        self.assertEqual(result['data'][0]['last_modified'], tstamp)

    def test_default_sort_is_last_modified_records_have_same_status(self):
        self.resource.request.validated['querystring'] = {'_sort': ['status']}
        result = self.resource.collection_get()
        self.assertEqual(result['data'][0]['status'], 0)
        self.assertEqual(result['data'][1]['status'], 0)
        self.assertGreater(result['data'][0]['last_modified'],
                           result['data'][1]['last_modified'])


class SubobjectSortingTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()

        indices = list(range(20))
        random.shuffle(indices)

        for i in indices:
            record = {
                'party': {'candidate': 'Marie', 'voters': i},
                'location': 'Creuse'
            }
            self.model.create_record(record)

    def test_records_can_be_sorted_by_subobjects(self):
        self.resource.request.validated['querystring'] = {'_sort': ['party.voters']}
        result = self.resource.collection_get()
        values = [item['party']['voters'] for item in result['data']]
        self.assertEqual(sorted(values), values)

    def test_subobjects_sorting_is_ignored_if_not_object(self):
        self.resource.request.validated['querystring'] = {'_sort': ['location.city']}
        result = self.resource.collection_get()
        values = [item['party']['voters'] for item in result['data']]
        self.assertNotEqual(sorted(values), values)



import colander

from kinto.core import resource, utils
from kinto.core.events import ResourceChanged, ACTIONS
from pyramid.events import subscriber


def validate_member(node, member):
    if member.startswith('/buckets/') or member == 'system.Everyone':
        raise colander.Invalid(node, "'{}' is not a valid user ID.".format(member))


class GroupSchema(resource.ResourceSchema):
    members = colander.SchemaNode(colander.Sequence(),
                                  colander.SchemaNode(colander.String(),
                                                      validator=validate_member))


@resource.register(name='group',
                   collection_path='/buckets/{{bucket_id}}/groups',
                   record_path='/buckets/{{bucket_id}}/groups/{{id}}')
class Group(resource.ShareableResource):
    schema = GroupSchema

    def get_parent_id(self, request):
        bucket_id = request.matchdict['bucket_id']
        parent_id = utils.instance_uri(request, 'bucket', id=bucket_id)
        return parent_id


@subscriber(ResourceChanged,
            for_resources=('group',),
            for_actions=(ACTIONS.DELETE,))
def on_groups_deleted(event):
    """Some groups were deleted, remove them from users principals.
    """
    permission_backend = event.request.registry.permission

    for change in event.impacted_records:
        group = change['old']
        bucket_id = event.payload['bucket_id']
        group_uri = utils.instance_uri(event.request, 'group',
                                       bucket_id=bucket_id,
                                       id=group['id'])

        permission_backend.remove_principal(group_uri)


@subscriber(ResourceChanged,
            for_resources=('group',),
            for_actions=(ACTIONS.CREATE, ACTIONS.UPDATE))
def on_groups_changed(event):
    """Some groups were changed, update users principals.
    """
    permission_backend = event.request.registry.permission

    for change in event.impacted_records:
        if 'old' in change:
            existing_record_members = set(change['old'].get('members', []))
        else:
            existing_record_members = set()

        group = change['new']
        group_uri = '/buckets/{bucket_id}/groups/{id}'.format(id=group['id'],
                                                              **event.payload)
        new_record_members = set(group.get('members', []))
        new_members = new_record_members - existing_record_members
        removed_members = existing_record_members - new_record_members

        for member in new_members:
            # Add the group to the member principal.
            permission_backend.add_user_principal(member, group_uri)

        for member in removed_members:
            # Remove the group from the member principal.
            permission_backend.remove_user_principal(member, group_uri)



from pyramid import authentication as base_auth

from kinto.core import utils
from kinto.core.api import OpenAPI


class BasicAuthAuthenticationPolicy(base_auth.BasicAuthAuthenticationPolicy):
    """Basic auth implementation.

    Allow any user with any credentials (e.g. there is no need to create an
    account).

    """
    def __init__(self, *args, **kwargs):
        def noop_check(*a):
            return []
        super().__init__(noop_check, *args, **kwargs)

    def effective_principals(self, request):
        # Bypass default Pyramid construction of principals because
        # Pyramid multiauth already adds userid, Authenticated and Everyone
        # principals.
        return []

    def unauthenticated_userid(self, request):
        settings = request.registry.settings

        credentials = base_auth.extract_http_basic_credentials(request)
        if credentials:
            username, password = credentials
            if not username:
                return

            hmac_secret = settings['userid_hmac_secret']
            credentials = '{}:{}'.format(*credentials)
            userid = utils.hmac_digest(hmac_secret, credentials)
            return userid


def includeme(config):
    config.add_api_capability(
        "basicauth",
        description="Very basic authentication sessions. Not for production use.",
        url="http://kinto.readthedocs.io/en/stable/api/1.x/authentication.html",
    )
    OpenAPI.expose_authentication_method('basicauth', {'type': 'basic'})



import colander

from kinto.core import resource
from kinto.core.utils import instance_uri
from kinto.core.storage import Filter
from kinto.core.resource.viewset import ViewSet


class HistorySchema(resource.ResourceSchema):
    user_id = colander.SchemaNode(colander.String())
    uri = colander.SchemaNode(colander.String())
    action = colander.SchemaNode(colander.String())
    date = colander.SchemaNode(colander.String())
    resource_name = colander.SchemaNode(colander.String())
    bucket_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    collection_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    group_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    record_id = colander.SchemaNode(colander.String(), missing=colander.drop)
    target = colander.SchemaNode(colander.Mapping())

    class Options:
        preserve_unknown = False


# Add custom OpenAPI tags/operation ids
collection_get_arguments = getattr(ViewSet, "collection_get_arguments", {})
collection_delete_arguments = getattr(ViewSet, "collection_delete_arguments", {})

get_history_arguments = {'tags': ['History'], 'operation_id': 'get_history',
                         **collection_get_arguments}
delete_history_arguments = {'tags': ['History'], 'operation_id': 'delete_history',
                            **collection_delete_arguments}


@resource.register(name='history',
                   collection_path='/buckets/{{bucket_id}}/history',
                   record_path=None,
                   collection_methods=('GET', 'DELETE'),
                   default_arguments={'tags': ['History']},
                   collection_get_arguments=get_history_arguments,
                   collection_delete_arguments=delete_history_arguments)
class History(resource.ShareableResource):

    schema = HistorySchema

    def get_parent_id(self, request):
        self.bucket_id = request.matchdict['bucket_id']
        return instance_uri(request, 'bucket', id=self.bucket_id)

    def _extract_filters(self):
        filters = super()._extract_filters()
        filters_str_id = []
        for filt in filters:
            if filt.field in ('record_id', 'collection_id', 'bucket_id'):
                if isinstance(filt.value, int):
                    filt = Filter(filt.field, str(filt.value), filt.operator)
            filters_str_id.append(filt)

        return filters_str_id



import uuid

from pyramid import httpexceptions
from pyramid.settings import asbool
from pyramid.security import NO_PERMISSION_REQUIRED, Authenticated

from kinto.core.errors import raise_invalid
from kinto.core.events import ACTIONS
from kinto.core.utils import (
    build_request, reapply_cors, hmac_digest, instance_uri, view_lookup)

from kinto.authorization import RouteFactory
from kinto.views.buckets import Bucket
from kinto.views.collections import Collection


def create_bucket(request, bucket_id):
    """Create a bucket if it doesn't exists."""
    bucket_put = (request.method.lower() == 'put' and
                  request.path.endswith('buckets/default'))
    # Do nothing if current request will already create the bucket.
    if bucket_put:
        return

    # Do not intent to create multiple times per request (e.g. in batch).
    already_created = request.bound_data.setdefault('buckets', {})
    if bucket_id in already_created:
        return

    bucket_uri = instance_uri(request, 'bucket', id=bucket_id)
    bucket = resource_create_object(request=request,
                                    resource_cls=Bucket,
                                    uri=bucket_uri)
    already_created[bucket_id] = bucket


def create_collection(request, bucket_id):
    # Do nothing if current request does not involve a collection.
    subpath = request.matchdict.get('subpath')
    if not (subpath and subpath.rstrip('/').startswith('collections/')):
        return

    collection_id = subpath.split('/')[1]
    collection_uri = instance_uri(request, 'collection',
                                  bucket_id=bucket_id,
                                  id=collection_id)

    # Do not intent to create multiple times per request (e.g. in batch).
    already_created = request.bound_data.setdefault('collections', {})
    if collection_uri in already_created:
        return

    # Do nothing if current request will already create the collection.
    collection_put = (request.method.lower() == 'put' and
                      request.path.endswith(collection_id))
    if collection_put:
        return

    collection = resource_create_object(request=request,
                                        resource_cls=Collection,
                                        uri=collection_uri)
    already_created[collection_uri] = collection


def resource_create_object(request, resource_cls, uri):
    """In the default bucket, the bucket and collection are implicitly
    created. This helper instantiate the resource and simulate a request
    with its RootFactory on the instantiated resource.
    :returns: the created object
    :rtype: dict
    """
    resource_name, matchdict = view_lookup(request, uri)

    # Build a fake request, mainly used to populate the create events that
    # will be triggered by the resource.
    fakerequest = build_request(request, {
        'method': 'PUT',
        'path': uri,
    })
    fakerequest.matchdict = matchdict
    fakerequest.bound_data = request.bound_data
    fakerequest.authn_type = request.authn_type
    fakerequest.selected_userid = request.selected_userid
    fakerequest.errors = request.errors
    fakerequest.current_resource_name = resource_name

    obj_id = matchdict['id']

    # Fake context, required to instantiate a resource.
    context = RouteFactory(fakerequest)
    context.resource_name = resource_name
    resource = resource_cls(fakerequest, context)

    # Check that provided id is valid for this resource.
    if not resource.model.id_generator.match(obj_id):
        error_details = {
            'location': 'path',
            'description': "Invalid {} id".format(resource_name)
        }
        raise_invalid(resource.request, **error_details)

    data = {'id': obj_id}
    obj = resource.model.create_record(data, ignore_conflict=True)
    # Since the current request is not a resource (but a straight Service),
    # we simulate a request on a resource.
    # This will be used in the resource event payload.
    resource.postprocess(obj, action=ACTIONS.CREATE)
    return obj


def default_bucket(request):
    if request.method.lower() == 'options':
        path = request.path.replace('default', 'unknown')
        subrequest = build_request(request, {
            'method': 'OPTIONS',
            'path': path
        })
        return request.invoke_subrequest(subrequest)

    if Authenticated not in request.effective_principals:
        # Pass through the forbidden_view_config
        raise httpexceptions.HTTPForbidden()

    settings = request.registry.settings

    if asbool(settings['readonly']):
        raise httpexceptions.HTTPMethodNotAllowed()

    bucket_id = request.default_bucket_id

    # Implicit object creations.
    # Make sure bucket exists
    create_bucket(request, bucket_id)
    # Make sure the collection exists
    create_collection(request, bucket_id)

    path = request.path.replace('/buckets/default', '/buckets/{}'.format(bucket_id))
    querystring = request.url[(request.url.index(request.path) +
                               len(request.path)):]
    try:
        # If 'id' is provided as 'default', replace with actual bucket id.
        body = request.json
        body['data']['id'] = body['data']['id'].replace('default', bucket_id)
    except:
        body = request.body or {"data": {}}
    subrequest = build_request(request, {
        'method': request.method,
        'path': path + querystring,
        'body': body,
    })
    subrequest.bound_data = request.bound_data

    try:
        response = request.invoke_subrequest(subrequest)
    except httpexceptions.HTTPException as error:
        is_redirect = error.status_code < 400
        if error.content_type == 'application/json' or is_redirect:
            response = reapply_cors(subrequest, error)
        else:
            # Ask the upper level to format the error.
            raise error
    return response


def default_bucket_id(request):
    settings = request.registry.settings
    secret = settings['userid_hmac_secret']
    # Build the user unguessable bucket_id UUID from its user_id
    digest = hmac_digest(secret, request.prefixed_userid)
    return str(uuid.UUID(digest[:32]))


def get_user_info(request):
    user_info = {
        'id': request.prefixed_userid,
        'bucket': request.default_bucket_id
    }
    return user_info


def includeme(config):
    # Redirect default to the right endpoint
    config.add_view(default_bucket,
                    route_name='default_bucket',
                    permission=NO_PERMISSION_REQUIRED)
    config.add_view(default_bucket,
                    route_name='default_bucket_collection',
                    permission=NO_PERMISSION_REQUIRED)

    config.add_route('default_bucket_collection',
                     '/buckets/default/{subpath:.*}')
    config.add_route('default_bucket', '/buckets/default')

    # Provide helpers
    config.add_request_method(default_bucket_id, reify=True)
    # Override kinto.core default user info
    config.add_request_method(get_user_info)

    config.add_api_capability(
        "default_bucket",
        description="The default bucket is an alias for a personal"
                    " bucket where collections are created implicitly.",
        url="https://kinto.readthedocs.io/en/latest/api/1.x/"
            "buckets.html#personal-bucket-default")



import mock

from kinto.core import scripts
from kinto.core.storage.exceptions import RecordNotFoundError
from kinto.core.testing import unittest


class InitSchemaTest(unittest.TestCase):
    def setUp(self):
        self.registry = mock.MagicMock()

    def test_migrate_calls_initialize_schema_on_backends(self):
        scripts.migrate({'registry': self.registry})
        self.assertTrue(self.registry.storage.initialize_schema.called)
        self.assertTrue(self.registry.cache.initialize_schema.called)
        self.assertTrue(self.registry.permission.initialize_schema.called)

    def test_migrate_skips_missing_backends(self):
        class FakeRegistry:
            settings = dict()
            storage = mock.MagicMock()
        registry = FakeRegistry()
        scripts.migrate({'registry': registry})
        self.assertTrue(registry.storage.initialize_schema.called)

    def test_migrate_in_read_only_display_an_error(self):
        with mock.patch('kinto.core.scripts.logger') as mocked:
            self.registry.settings = {'readonly': 'true'}
            scripts.migrate({'registry': self.registry})
            mocked.error.assert_any_call('Cannot migrate the storage backend '
                                         'while in readonly mode.')
            mocked.error.assert_any_call('Cannot migrate the permission '
                                         'backend while in readonly mode.')

    def test_migrate_in_dry_run_mode(self):
        scripts.migrate({'registry': self.registry}, dry_run=True)
        reg = self.registry
        reg.storage.initialize_schema.assert_called_with(dry_run=True)
        reg.cache.initialize_schema.assert_called_with(dry_run=True)
        reg.permission.initialize_schema.assert_called_with(dry_run=True)


class DeleteCollectionTest(unittest.TestCase):
    def setUp(self):
        self.registry = mock.MagicMock()

    def test_delete_collection_in_read_only_display_an_error(self):
        with mock.patch('kinto.core.scripts.logger') as mocked:
            self.registry.settings = {'readonly': 'true'}
            code = scripts.delete_collection({'registry': self.registry},
                                             'test_bucket',
                                             'test_collection')
            assert code == 31
            mocked.error.assert_any_call('Cannot delete the collection while '
                                         'in readonly mode.')

    def test_delete_collection_remove_collection_records(self):
        self.registry.storage.delete_all.return_value = [
            {"id": "1234"}, {"id": "5678"}
        ]

        with mock.patch('kinto.core.scripts.logger') as mocked:
            scripts.delete_collection({'registry': self.registry},
                                      'test_bucket',
                                      'test_collection')

        self.registry.storage.delete_all.assert_called_with(
            collection_id='record',
            parent_id='/buckets/test_bucket/collections/test_collection',
            with_deleted=False)
        self.registry.storage.delete.assert_called_with(
            collection_id='collection',
            parent_id='/buckets/test_bucket',
            object_id='test_collection',
            with_deleted=False)
        self.registry.permission.delete_object_permissions.assert_called_with(
            '/buckets/test_bucket/collections/test_collection',
            '/buckets/test_bucket/collections/test_collection/records/1234',
            '/buckets/test_bucket/collections/test_collection/records/5678')

        mocked.info.assert_any_call('2 record(s) were deleted.')
        mocked.info.assert_any_call(
            "'/buckets/test_bucket/collections/test_collection' "
            "collection object was deleted.")

    def test_delete_collection_tell_when_no_records_where_found(self):
        self.registry.storage.delete_all.return_value = []

        with mock.patch('kinto.core.scripts.logger') as mocked:
            scripts.delete_collection({'registry': self.registry},
                                      'test_bucket',
                                      'test_collection')

        mocked.info.assert_any_call(
            "No records found for "
            "'/buckets/test_bucket/collections/test_collection'.")
        mocked.info.assert_any_call(
            "'/buckets/test_bucket/collections/test_collection' "
            "collection object was deleted.")
        mocked.info.assert_any_call("Related permissions were deleted.")

    def test_delete_collection_raise_if_the_bucket_does_not_exist(self):
        self.registry.storage.get.side_effect = RecordNotFoundError
        with mock.patch('kinto.core.scripts.logger') as mocked:
            resp = scripts.delete_collection({'registry': self.registry},
                                             'test_bucket',
                                             'test_collection')
        assert resp == 32
        mocked.error.assert_called_with(
            "Bucket '/buckets/test_bucket' does not exist.")

    def test_delete_collection_raise_if_the_collection_does_not_exist(self):
        self.registry.storage.get.side_effect = ['', RecordNotFoundError]
        with mock.patch('kinto.core.scripts.logger') as mocked:
            resp = scripts.delete_collection({'registry': self.registry},
                                             'test_bucket',
                                             'test_collection')
        assert resp == 33
        mocked.error.assert_called_with(
            "Collection '/buckets/test_bucket/collections/test_collection' "
            "does not exist.")



import os
import threading
import unittest
from collections import defaultdict

import mock
import webtest
from cornice import errors as cornice_errors
from pyramid.url import parse_url_overrides

from kinto.core import DEFAULT_SETTINGS
from kinto.core import statsd
from kinto.core.storage import generators
from kinto.core.utils import sqlalchemy, follow_subrequest, encode64

skip_if_travis = unittest.skipIf('TRAVIS' in os.environ, "travis")
skip_if_no_postgresql = unittest.skipIf(sqlalchemy is None, "postgresql is not installed.")
skip_if_no_statsd = unittest.skipIf(not statsd.statsd_module, "statsd is not installed.")


class DummyRequest(mock.MagicMock):
    """Fully mocked request.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.upath_info = '/v0/'
        self.registry = mock.MagicMock(settings={**DEFAULT_SETTINGS})
        self.registry.id_generators = defaultdict(generators.UUID4)
        self.GET = {}
        self.headers = {}
        self.errors = cornice_errors.Errors()
        self.authenticated_userid = 'bob'
        self.authn_type = 'basicauth'
        self.prefixed_userid = 'basicauth:bob'
        self.effective_principals = [
            'system.Everyone',
            'system.Authenticated',
            'bob']
        self.prefixed_principals = self.effective_principals + [self.prefixed_userid]
        self.json = {}
        self.validated = {}
        self.matchdict = {}
        self.response = mock.MagicMock(headers={})

        def route_url(*a, **kw):
            # XXX: refactor DummyRequest to take advantage of `pyramid.testing`
            parts = parse_url_overrides(kw)
            return ''.join([p for p in parts if p])

        self.route_url = route_url

    follow_subrequest = follow_subrequest


def get_request_class(prefix):

    class PrefixedRequestClass(webtest.app.TestRequest):

        @classmethod
        def blank(cls, path, *args, **kwargs):
            if prefix:
                path = '/{prefix}{path}'.format(prefix=prefix, path=path)
            return webtest.app.TestRequest.blank(path, *args, **kwargs)

    return PrefixedRequestClass


class FormattedErrorMixin:
    """Test mixin in order to perform advanced error responses assertions.
    """

    def assertFormattedError(self, response, code, errno, error,
                             message=None, info=None):
        self.assertIn('application/json', response.headers['Content-Type'])
        self.assertEqual(response.json['code'], code)
        self.assertEqual(response.json['errno'], errno.value)
        self.assertEqual(response.json['error'], error)
        if message is not None:
            self.assertIn(message, response.json['message'])
        else:  # pragma: no cover
            self.assertNotIn('message', response.json)

        if info is not None:
            self.assertIn(info, response.json['info'])
        else:  # pragma: no cover
            self.assertNotIn('info', response.json)


def get_user_headers(user):
    """Helper to obtain a Basic Auth authorization headers from the specified
    `user` (e.g. ``"user:pass"``)

    :rtype: dict
    """
    credentials = "{}:secret".format(user)
    authorization = 'Basic {}'.format(encode64(credentials))
    return {
        'Authorization': authorization
    }


class BaseWebTest:
    """Base Web Test to test your kinto.core service.

    It setups the database before each test and delete it after.
    """

    api_prefix = "v0"
    """URL version prefix"""

    entry_point = None
    """Main application entry"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.app = self.make_app()
        self.storage = self.app.app.registry.storage
        self.cache = self.app.app.registry.cache
        self.permission = self.app.app.registry.permission

        self.storage.initialize_schema()
        self.permission.initialize_schema()
        self.cache.initialize_schema()

        self.headers = {
            'Content-Type': 'application/json'
        }

    def make_app(self, settings=None, config=None):
        """Instantiate the application and setup requests to use the api
        prefix.

        :param dict settings: extra settings values
        :param pyramid.config.Configurator config: already initialized config
        :returns: webtest application instance
        """
        settings = self.get_app_settings(extras=settings)

        try:
            main = self.entry_point.__func__
        except AttributeError:  # pragma: no cover
            main = self.entry_point.im_func

        wsgi_app = main({}, config=config, **settings)
        app = webtest.TestApp(wsgi_app)
        app.RequestClass = get_request_class(self.api_prefix)
        return app

    def get_app_settings(self, extras=None):
        """Application settings to be used. Override to tweak default settings
        for the tests.

        :param dict extras: extra settings values
        :rtype: dict
        """
        settings = {**DEFAULT_SETTINGS}

        settings['storage_backend'] = 'kinto.core.storage.memory'
        settings['cache_backend'] = 'kinto.core.cache.memory'
        settings['permission_backend'] = 'kinto.core.permission.memory'

        settings.update(extras or None)

        return settings

    def tearDown(self):
        super().tearDown()
        self.storage.flush()
        self.cache.flush()
        self.permission.flush()


class ThreadMixin:

    def setUp(self):
        super().setUp()
        self._threads = []

    def tearDown(self):
        super().tearDown()

        for thread in self._threads:
            thread.join()

    def _create_thread(self, *args, **kwargs):
        thread = threading.Thread(*args, **kwargs)
        self._threads.append(thread)
        return thread



from .support import SwaggerTest


class SwaggerPluginsTest(SwaggerTest):

    def test_get_history(self):
        op = self.resources['History'].get_history
        self.request.path = {
            'bucket_id': 'b1',
        }
        self.validate_request_call(op)

    def test_delete_history(self):
        op = self.resources['History'].delete_history
        self.request.path = {
            'bucket_id': 'b1',
        }
        self.validate_request_call(op)



import mock

import colander
from pyramid import httpexceptions

from kinto.core.resource import ResourceSchema
from kinto.core.errors import ERRORS

from . import BaseTest


class GetTest(BaseTest):
    def test_get_record_returns_all_fields(self):
        record = self.model.create_record({'field': 'value'})
        self.resource.record_id = record['id']
        result = self.resource.get()['data']
        self.assertIn(self.resource.model.id_field, result)
        self.assertIn(self.resource.model.modified_field, result)
        self.assertIn('field', result)

    def test_etag_is_provided(self):
        record = self.model.create_record({'field': 'value'})
        self.resource.record_id = record['id']
        self.resource.get()
        self.assertIn('ETag', self.last_response.headers)

    def test_etag_contains_record_timestamp(self):
        record = self.model.create_record({'field': 'value'})
        self.resource.record_id = record['id']
        # Create another one, bump collection timestamp.
        self.model.create_record({'field': 'value'})
        self.resource.get()
        expected = '"{}"'.format(record['last_modified'])
        self.assertEqual(expected, self.last_response.headers['ETag'])


class PutTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.record = self.model.create_record({'field': 'old'})
        self.resource.record_id = self.record['id']

    def test_etag_is_provided(self):
        self.validated['body'] = {'data': {'field': 'new'}}
        self.resource.put()
        self.assertIn('ETag', self.last_response.headers)

    def test_etag_contains_record_new_timestamp(self):
        self.validated['body'] = {'data': {'field': 'new'}}
        new = self.resource.put()['data']
        expected = '"{}"'.format(new['last_modified'])
        self.assertEqual(expected, self.last_response.headers['ETag'])

    def test_returns_201_if_created(self):
        self.resource.record_id = self.resource.model.id_generator()
        self.validated['body'] = {'data': {'field': 'new'}}
        self.resource.put()
        self.assertEqual(self.last_response.status_code, 201)

    def test_relies_on_collection_create(self):
        self.resource.record_id = self.resource.model.id_generator()
        self.validated['body'] = {'data': {'field': 'new'}}
        with mock.patch.object(self.model, 'create_record') as patched:
            self.resource.put()
            self.assertEqual(patched.call_count, 1)

    def test_relies_on_collection_create_even_when_previously_deleted(self):
        record = self.model.create_record({'field': 'value'})
        self.resource.record_id = record['id']
        self.resource.delete()['data']

        self.validated['body'] = {'data': {'field': 'new'}}
        with mock.patch.object(self.model, 'create_record') as patched:
            self.resource.put()
            self.assertEqual(patched.call_count, 1)

    def test_replace_record_returns_updated_fields(self):
        self.validated['body'] = {'data': {'field': 'new'}}
        result = self.resource.put()['data']
        self.assertEqual(self.record['id'], result['id'])
        self.assertNotEqual(self.record['last_modified'],
                            result['last_modified'])
        self.assertNotEqual(self.record['field'], 'new')

    def test_last_modified_is_kept_if_present(self):
        new_last_modified = self.record['last_modified'] + 20
        self.validated['body'] = {'data': {
            'field': 'new',
            'last_modified': new_last_modified}
        }
        result = self.resource.put()['data']
        self.assertEqual(result['last_modified'], new_last_modified)

    def test_last_modified_is_dropped_if_same_as_previous(self):
        self.validated['body'] = {'data': {
            'field': 'new',
            'last_modified': self.record['last_modified']}
        }
        result = self.resource.put()['data']
        self.assertGreater(result['last_modified'],
                           self.record['last_modified'])

    def test_last_modified_is_dropped_if_lesser_than_existing(self):
        new_last_modified = self.record['last_modified'] - 20
        self.validated['body'] = {'data': {
            'field': 'new',
            'last_modified': new_last_modified}
        }
        result = self.resource.put()['data']
        self.assertNotEqual(result['last_modified'],
                            self.record['last_modified'])

    def test_cannot_replace_with_different_id(self):
        self.validated['body'] = {'data': {'id': 'abc'}}
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.put)

    def test_last_modified_is_overwritten_on_replace(self):
        self.validated['body'] = {'data': {'last_modified': 123}}
        result = self.resource.put()['data']
        self.assertNotEqual(result['last_modified'], 123)

    def test_storage_is_not_used_if_context_provides_current_record(self):
        self.resource.context.current_record = {'id': 'hola'}
        self.validated['body'] = {'data': {}}
        with mock.patch.object(self.resource.model, 'get_record') as get:
            self.resource.put()
            self.assertFalse(get.called)

    def test_put_over_tombstone_create_new_record(self):
        self.resource.delete()
        result = self.resource.put()
        self.assertNotIn("old", result)
        self.assertNotIn("deleted", result)


class DeleteTest(BaseTest):
    def test_delete_record_returns_last_timestamp(self):
        record = {'field': 'value'}
        record = {**self.model.create_record(record)}
        self.resource.record_id = record['id']
        result = self.resource.delete()['data']
        self.assertNotEqual(result['last_modified'], record['last_modified'])

    def test_etag_is_provided(self):
        record = self.model.create_record({'field': 'value'})
        self.resource.record_id = record['id']
        self.resource.delete()
        self.assertIn('ETag', self.last_response.headers)

    def test_etag_contains_deleted_timestamp(self):
        record = self.model.create_record({'field': 'value'})
        self.resource.record_id = record['id']
        deleted = self.resource.delete()
        expected = '"{}"'.format(deleted['data']['last_modified'])
        self.assertEqual(expected, self.last_response.headers['ETag'])

    def test_delete_record_returns_stripped_record(self):
        record = self.model.create_record({'field': 'value'})
        self.resource.record_id = record['id']
        result = self.resource.delete()['data']
        self.assertEqual(result['id'], record['id'])
        self.assertNotIn('field', result)
        self.assertIn('last_modified', result)

    def test_delete_uses_last_modified_from_querystring(self):
        record = self.model.create_record({'field': 'value'})
        last_modified = record[self.model.modified_field] + 20
        self.resource.record_id = record['id']
        self.validated['querystring']['last_modified'] = last_modified
        result = self.resource.delete()['data']
        self.validated = self.validated
        self.assertEqual(result[self.model.modified_field], last_modified)
        self.validated['querystring'] = {
            '_since': 0, 'deleted': True}
        result = self.resource.collection_get()
        self.assertEqual(len(result['data']), 1)
        retrieved = result['data'][0]
        self.assertEqual(retrieved[self.model.modified_field], last_modified)

    def test_delete_ignores_last_modified_if_equal(self):
        record = self.model.create_record({'field': 'value'})
        last_modified = record[self.model.modified_field]
        self.resource.record_id = record['id']
        self.validated['querystring']['last_modified'] = last_modified
        result = self.resource.delete()['data']
        self.assertGreater(result[self.model.modified_field], last_modified)

    def test_delete_ignores_last_modified_if_less(self):
        record = self.model.create_record({'field': 'value'})
        last_modified = record[self.model.modified_field] - 20
        self.resource.record_id = record['id']
        self.validated['querystring']['last_modified'] = last_modified
        result = self.resource.delete()['data']
        self.assertGreater(result[self.model.modified_field], last_modified)


class PatchTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.stored = self.model.create_record({})
        self.resource.record_id = self.stored['id']
        self.validated['body'] = {'data': {'position': 10}}

        class ArticleSchema(ResourceSchema):
            unread = colander.SchemaNode(colander.Boolean(), missing=colander.drop)
            position = colander.SchemaNode(colander.Int(), missing=colander.drop)

        self.resource.schema = ArticleSchema

        self.result = self.resource.patch()['data']

    def test_etag_is_provided(self):
        self.assertIn('ETag', self.last_response.headers)

    def test_etag_contains_record_new_timestamp(self):
        expected = '"{}"'.format(self.result['last_modified'])
        self.assertEqual(expected, self.last_response.headers['ETag'])

    def test_etag_contains_old_timestamp_if_no_field_changed(self):
        self.validated['body'] = {'data': {'position': 10}}
        self.resource.patch()['data']
        expected = '"{}"'.format(self.result['last_modified'])
        self.assertEqual(expected, self.last_response.headers['ETag'])

    def test_modify_record_updates_timestamp(self):
        before = self.stored['last_modified']
        after = self.result['last_modified']
        self.assertNotEquals(after, before)

    def test_patch_record_returns_updated_fields(self):
        self.assertEquals(self.stored['id'], self.result['id'])
        self.assertEquals(self.result['position'], 10)

    def test_record_timestamp_is_not_updated_if_none_for_missing_field(self):
        self.validated['body'] = {'data': {'polo': None}}
        result = self.resource.patch()['data']
        self.assertEquals(self.result['last_modified'],
                          result['last_modified'])

    def test_record_timestamp_is_not_updated_if_no_field_changed(self):
        self.validated['body'] = {'data': {'position': 10}}
        result = self.resource.patch()['data']
        self.assertEquals(self.result['last_modified'],
                          result['last_modified'])

    def test_collection_timestamp_is_not_updated_if_no_field_changed(self):
        self.validated['body'] = {'data': {'position': 10}}
        self.resource.patch()
        self.resource = self.resource_class(request=self.get_request(),
                                            context=self.get_context())
        self.resource.request.validated = self.validated
        self.resource.collection_get()['data']
        last_modified = int(self.last_response.headers['ETag'][1:-1])
        self.assertEquals(self.result['last_modified'], last_modified)

    def test_timestamp_is_not_updated_if_no_change_after_preprocessed(self):
        with mock.patch.object(self.resource, 'process_record') as mocked:
            mocked.return_value = self.result
            self.validated['body'] = {'data': {'position': 20}}
            result = self.resource.patch()['data']
            self.assertEquals(self.result['last_modified'],
                              result['last_modified'])

    def test_returns_changed_fields_among_provided_if_behaviour_is_diff(self):
        self.validated['body'] = {'data': {'unread': True, 'position': 15}}
        self.validated['header'] = {
            'Response-Behavior': 'diff'
        }
        with mock.patch.object(self.resource.model, 'update_record',
                               return_value={'unread': True, 'position': 0}):
            result = self.resource.patch()['data']
        self.assertDictEqual(result, {'position': 0})

    def test_returns_changed_fields_if_behaviour_is_light(self):
        self.validated['body'] = {'data': {'unread': True, 'position': 15}}
        self.validated['header'] = {
            'Response-Behavior': 'light'
        }
        with mock.patch.object(self.resource.model, 'update_record',
                               return_value={'unread': True, 'position': 0}):
            result = self.resource.patch()['data']
        self.assertDictEqual(result, {'unread': True, 'position': 0})


class MergePatchTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.stored = self.model.create_record({})
        self.resource.record_id = self.stored['id']
        self.headers = self.resource.request.headers
        self.headers['Content-Type'] = 'application/merge-patch+json'

        class ArticleSchema(ResourceSchema):
            unread = colander.SchemaNode(colander.Boolean(), missing=colander.drop)
            position = colander.SchemaNode(colander.Int(), missing=colander.drop)

        self.resource.schema = ArticleSchema

    def test_merge_patch_updates_attributes_recursively(self):
        self.validated['body'] = {'data': {'a': {'b': 'bbb',
                                                 'c': 'ccc'}}}
        self.resource.patch()
        self.validated['body'] = {'data': {'a': {'b': 'aaa',
                                                 'c': None}}}
        result = self.resource.patch()['data']
        self.assertEqual(result['a']['b'], 'aaa')

    def test_merge_patch_removes_attribute_if_none(self):
        self.validated['body'] = {'data': {'field': 'aaa'}}
        self.resource.patch()
        self.validated['body'] = {'data': {'field': None}}
        result = self.resource.patch()['data']
        self.assertNotIn('field', result)
        result = self.resource.get()['data']
        self.assertNotIn('field', result)

    def test_merge_patch_removes_attributes_recursively_if_none(self):
        self.validated['body'] = {'data': {'a': {'b': 'aaa'}}}
        self.resource.patch()
        self.validated['body'] = {'data': {'a': {'b': None}}}
        result = self.resource.patch()['data']
        self.assertIn('a', result)
        self.assertNotIn('b', result['a'])
        self.validated['body'] = {'data': {'aa': {'bb': {'cc': None}}}}
        result = self.resource.patch()['data']
        self.assertIn('aa', result)
        self.assertIn('bb', result['aa'])
        self.assertNotIn('cc', result['aa']['bb'])

    def test_merge_patch_doesnt_remove_attribute_if_false(self):
        self.validated['body'] = {'data': {'field': 0}}
        result = self.resource.patch()['data']
        self.assertIn('field', result)
        self.validated['body'] = {'data': {'field': False}}
        result = self.resource.patch()['data']
        self.assertIn('field', result)
        self.validated['body'] = {'data': {'field': {}}}
        result = self.resource.patch()['data']
        self.assertIn('field', result)

    def test_patch_doesnt_remove_attribute_if_not_merge_header(self):
        self.headers['Content-Type'] = 'application/json'
        self.validated['body'] = {'data': {'field': 'aaa'}}
        self.resource.patch()
        self.validated['body'] = {'data': {'field': None}}
        result = self.resource.patch()['data']
        self.assertIn('field', result)
        result = self.resource.get()['data']
        self.assertIn('field', result)

    def test_merge_patch_doesnt_remove_previously_inserted_nones(self):
        self.headers['Content-Type'] = 'application/json'
        self.validated['body'] = {'data': {'field': 'aaa'}}
        result = self.resource.patch()['data']
        self.validated['body'] = {'data': {'field': None}}
        result = self.resource.patch()['data']
        self.assertIn('field', result)
        self.headers['Content-Type'] = 'application/merge-patch+json'
        self.validated['body'] = {'data': {'position': 10}}
        result = self.resource.patch()['data']
        self.assertIn('field', result)


class JsonPatchTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.stored = self.model.create_record({})
        self.resource.record_id = self.stored['id']
        self.validated['body'] = {'data': {'a': 'aaa', 'b': ['bb', 'bbb'], 'd': []}}
        self.resource.schema = ResourceSchema
        self.result = self.resource.patch()['data']
        header = self.resource.request.headers
        header['Content-Type'] = 'application/json-patch+json'
        self.resource._is_json_patch = True

    def test_json_patch_add(self):
        self.validated['body'] = [
            {'op': 'add', 'path': '/data/c', 'value': 'ccc'},
            {'op': 'add', 'path': '/data/b/1', 'value': 'ddd'},
            {'op': 'add', 'path': '/data/b/-', 'value': 'eee'},
        ]
        result = self.resource.patch()['data']
        self.assertEqual(result['c'], 'ccc')
        self.assertEqual(result['b'][0], 'bb')
        self.assertEqual(result['b'][1], 'ddd')
        self.assertEqual(result['b'][2], 'bbb')
        self.assertEqual(result['b'][3], 'eee')
        self.assertEqual(len(result['b']), 4)

    def test_json_patch_remove(self):
        self.validated['body'] = [
            {'op': 'remove', 'path': '/data/a'},
            {'op': 'remove', 'path': '/data/b/0'},
        ]
        result = self.resource.patch()['data']
        self.assertNotIn('a', result)
        self.assertEqual(len(result['b']), 1)
        self.assertEqual(result['b'][0], 'bbb')

    def test_json_patch_test_success(self):
        self.validated['body'] = [
            {'op': 'test', 'path': '/data/a', 'value': 'aaa'},
        ]
        self.resource.patch()['data']

    def test_json_patch_test_failure(self):
        self.validated['body'] = [
            {'op': 'test', 'path': '/data/a', 'value': 'bbb'},
        ]
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.patch)

    def test_json_patch_move(self):
        self.validated['body'] = [
            {'op': 'move', 'from': '/data/a', 'path': '/data/c'},
            {'op': 'move', 'from': '/data/b/1', 'path': '/data/d/0'},
            {'op': 'move', 'from': '/data/b/0', 'path': '/data/e'},
        ]
        result = self.resource.patch()['data']
        self.assertNotIn('a', result)
        self.assertEqual(result['c'], 'aaa')
        self.assertEqual(len(result['b']), 0)
        self.assertEqual(result['d'][0], 'bbb')
        self.assertEqual(result['e'], 'bb')

    def test_json_patch_copy(self):
        self.validated['body'] = [
            {'op': 'copy', 'from': '/data/a', 'path': '/data/c'},
            {'op': 'copy', 'from': '/data/b/1', 'path': '/data/d/0'},
            {'op': 'copy', 'from': '/data/d/0', 'path': '/data/e'},
        ]
        result = self.resource.patch()['data']
        self.assertEqual(result['a'], 'aaa')
        self.assertEqual(result['c'], 'aaa')
        self.assertEqual(len(result['b']), 2)
        self.assertEqual(result['d'][0], 'bbb')
        self.assertEqual(result['e'], 'bbb')

    def test_json_patch_replace(self):
        self.validated['body'] = [
            {'op': 'replace', 'path': '/data/a', 'value': 'bbb'},
            {'op': 'replace', 'path': '/data/b/0', 'value': 'aaa'},
        ]
        result = self.resource.patch()['data']
        self.assertEqual(result['a'], 'bbb')
        self.assertEqual(result['b'][0], 'aaa')

    def test_json_patch_raises_400_on_invalid_path(self):
        self.validated['body'] = [
            {'op': 'remove', 'path': '/data/f'},
        ]
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.patch)
        self.validated['body'] = [
            {'op': 'move', 'from': '/data/f', 'path': '/data/what'},
        ]
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.patch)
        self.validated['body'] = [
            {'op': 'copy', 'from': '/data/what', 'path': '/data/f'},
        ]
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.patch)
        self.validated['body'] = [
            {'op': 'replace', 'path': '/data/c', 'value': 'ccc'},
        ]
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.patch)

    def test_json_patch_format_not_accepted_without_header(self):
        header = self.resource.request.headers
        header['Content-Type'] = 'application/json'
        self.resource._is_json_patch = False
        self.validated['body'] = [
            {'op': 'add', 'from': '/data/a', 'value': 'aaa'},
        ]
        self.assertRaises(AttributeError, self.resource.patch)


class UnknownRecordTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.unknown_id = '1cea99eb-5e3d-44ad-a53a-2fb68473b538'
        self.resource.record_id = self.unknown_id
        self.validated['body'] = {'data': {'field': 'new'}}

    def test_get_record_unknown_raises_404(self):
        self.assertRaises(httpexceptions.HTTPNotFound, self.resource.get)

    def test_modify_record_unknown_raises_404(self):
        self.assertRaises(httpexceptions.HTTPNotFound, self.resource.patch)

    def test_replace_record_unknown_creates_it(self):
        self.resource.put()
        self.model.get_record(self.unknown_id)

    def test_delete_record_unknown_raises_404(self):
        self.assertRaises(httpexceptions.HTTPNotFound, self.resource.delete)


class InvalidIdTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.resource.record_id = 'a*b'

    def test_get_with_invalid_id_raises_400(self):
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.get)

    def test_patch_with_invalid_id_raises_400(self):
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.patch)

    def test_put_with_invalid_id_raises_400(self):
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.put)

    def test_delete_with_invalid_id_raises_400(self):
        self.assertRaises(httpexceptions.HTTPBadRequest, self.resource.delete)


class ReadonlyFieldsTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.stored = self.model.create_record({'age': 32})
        self.resource.schema.Options.readonly_fields = ('age',)
        self.resource.record_id = self.stored['id']

    def assertReadonlyError(self, field):
        error = None
        try:
            self.resource.patch()
        except httpexceptions.HTTPBadRequest as e:
            error = e
        self.assertEqual(error.json, {
            'errno': ERRORS.INVALID_PARAMETERS.value,
            'message': 'Cannot modify {0}'.format(field),
            'code': 400,
            'error': 'Invalid parameters',
            'details': [{'description': 'Cannot modify age',
                         'location': 'body',
                         'name': 'age'}]})

    def test_can_specify_readonly_fields_if_not_changed(self):
        self.validated['body'] = {'data': {'age': self.stored['age']}}
        self.resource.patch()  # not raising

    def test_cannot_modify_readonly_field(self):
        self.validated['body'] = {'data': {'age': 16}}
        self.assertReadonlyError('age')



import os
import warnings
from collections import defaultdict

from kinto.core import logger
from kinto.core.storage import (
    StorageBase, exceptions,
    DEFAULT_ID_FIELD, DEFAULT_MODIFIED_FIELD, DEFAULT_DELETED_FIELD)
from kinto.core.storage.postgresql.client import create_from_config
from kinto.core.utils import COMPARISON, json, is_numeric


class Storage(StorageBase):
    """Storage backend using PostgreSQL.

    Recommended in production (*requires PostgreSQL 9.4 or higher*).

    Enable in configuration::

        kinto.storage_backend = kinto.core.storage.postgresql

    Database location URI can be customized::

        kinto.storage_url = postgres://user:pass@db.server.lan:5432/dbname

    Alternatively, username and password could also rely on system user ident
    or even specified in :file:`~/.pgpass` (*see PostgreSQL documentation*).

    .. note::

        Some tables and indices are created when ``kinto migrate`` is run.
        This requires some privileges on the database, or some error will
        be raised.

        **Alternatively**, the schema can be initialized outside the
        python application, using the SQL file located in
        :file:`kinto/core/storage/postgresql/schema.sql`. This allows to
        distinguish schema manipulation privileges from schema usage.


    A connection pool is enabled by default::

        kinto.storage_pool_size = 10
        kinto.storage_maxoverflow = 10
        kinto.storage_max_backlog = -1
        kinto.storage_pool_recycle = -1
        kinto.storage_pool_timeout = 30
        kinto.cache_poolclass =
            kinto.core.storage.postgresql.pool.QueuePoolWithMaxBacklog

    The ``max_backlog``  limits the number of threads that can be in the queue
    waiting for a connection.  Once this limit has been reached, any further
    attempts to acquire a connection will be rejected immediately, instead of
    locking up all threads by keeping them waiting in the queue.

    See `dedicated section in SQLAlchemy documentation
    <http://docs.sqlalchemy.org/en/rel_1_0/core/engines.html>`_
    for default values and behaviour.

    .. note::

        Using a `dedicated connection pool <http://pgpool.net>`_ is still
        recommended to allow load balancing, replication or limit the number
        of connections used in a multi-process deployment.

    """  # NOQA

    schema_version = 15

    def __init__(self, client, max_fetch_size, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.client = client
        self._max_fetch_size = max_fetch_size

    def _execute_sql_file(self, filepath):
        with open(filepath) as f:
            schema = f.read()
        # Since called outside request, force commit.
        with self.client.connect(force_commit=True) as conn:
            conn.execute(schema)

    def initialize_schema(self, dry_run=False):
        """Create PostgreSQL tables, and run necessary schema migrations.

        .. note::

            Relies on JSONB fields, available in recent versions of PostgreSQL.
        """
        here = os.path.abspath(os.path.dirname(__file__))

        version = self._get_installed_version()
        if not version:
            filepath = os.path.join(here, 'schema.sql')
            logger.info("Create PostgreSQL storage schema at version "
                        "{} from {}".format(self.schema_version, filepath))
            # Create full schema.
            self._check_database_encoding()
            self._check_database_timezone()
            # Create full schema.
            if not dry_run:
                self._execute_sql_file(filepath)
                logger.info('Created PostgreSQL storage schema (version {}).'.format(
                    self.schema_version))
            return

        logger.info('Detected PostgreSQL storage schema version {}.'.format(version))
        migrations = [(v, v + 1) for v in range(version, self.schema_version)]
        if not migrations:
            logger.info('PostgreSQL storage schema is up-to-date.')
            return

        for migration in migrations:
            # Check order of migrations.
            expected = migration[0]
            current = self._get_installed_version()
            error_msg = "Expected version {}. Found version {}."
            if not dry_run and expected != current:
                raise AssertionError(error_msg.format(expected, current))

            logger.info('Migrate PostgreSQL storage schema from'
                        ' version {} to {}.'.format(*migration))
            filename = 'migration_{0:03d}_{1:03d}.sql'.format(*migration)
            filepath = os.path.join(here, 'migrations', filename)
            logger.info("Execute PostgreSQL storage migration from {}".format(filepath))
            if not dry_run:
                self._execute_sql_file(filepath)
        logger.info("PostgreSQL storage schema migration {}".format(
            "simulated." if dry_run else "done."))

    def _check_database_timezone(self):
        # Make sure database has UTC timezone.
        query = "SELECT current_setting('TIMEZONE') AS timezone;"
        with self.client.connect() as conn:
            result = conn.execute(query)
            record = result.fetchone()
        timezone = record['timezone'].upper()
        if timezone != 'UTC':  # pragma: no cover
            msg = 'Database timezone is not UTC ({})'.format(timezone)
            warnings.warn(msg)
            logger.warning(msg)

    def _check_database_encoding(self):
        # Make sure database is UTF-8.
        query = """
        SELECT pg_encoding_to_char(encoding) AS encoding
          FROM pg_database
         WHERE datname =  current_database();
        """
        with self.client.connect() as conn:
            result = conn.execute(query)
            record = result.fetchone()
        encoding = record['encoding'].lower()
        if encoding != 'utf8':  # pragma: no cover
            raise AssertionError('Unexpected database encoding {}'.format(encoding))

    def _get_installed_version(self):
        """Return current version of schema or None if not any found.
        """
        query = "SELECT tablename FROM pg_tables WHERE tablename = 'metadata';"
        with self.client.connect() as conn:
            result = conn.execute(query)
            tables_exist = result.rowcount > 0

        if not tables_exist:
            return

        query = """
        SELECT value AS version
          FROM metadata
         WHERE name = 'storage_schema_version'
         ORDER BY LPAD(value, 3, '0') DESC;
        """
        with self.client.connect() as conn:
            result = conn.execute(query)
            if result.rowcount > 0:
                return int(result.fetchone()['version'])
            else:
                # Guess current version.
                query = "SELECT COUNT(*) FROM metadata;"
                result = conn.execute(query)
                was_flushed = int(result.fetchone()[0]) == 0
                if was_flushed:
                    error_msg = 'Missing schema history: consider version {}.'
                    logger.warning(error_msg.format(self.schema_version))
                    return self.schema_version

                # In the first versions of Cliquet, there was no migration.
                return 1

    def flush(self, auth=None):
        """Delete records from tables without destroying schema. Mainly used
        in tests suites.
        """
        query = """
        DELETE FROM deleted;
        DELETE FROM records;
        DELETE FROM timestamps;
        DELETE FROM metadata;
        """
        with self.client.connect(force_commit=True) as conn:
            conn.execute(query)
        logger.debug('Flushed PostgreSQL storage tables')

    def collection_timestamp(self, collection_id, parent_id, auth=None):
        query = """
        SELECT as_epoch(collection_timestamp(:parent_id, :collection_id))
            AS last_modified;
        """
        placeholders = dict(parent_id=parent_id, collection_id=collection_id)
        with self.client.connect(readonly=False) as conn:
            result = conn.execute(query, placeholders)
            record = result.fetchone()
        return record['last_modified']

    def create(self, collection_id, parent_id, record, id_generator=None,
               id_field=DEFAULT_ID_FIELD,
               modified_field=DEFAULT_MODIFIED_FIELD,
               auth=None, ignore_conflict=False):
        id_generator = id_generator or self.id_generator
        record = {**record}
        if id_field in record:
            # Raise unicity error if record with same id already exists.
            try:
                existing = self.get(collection_id, parent_id, record[id_field])
                if ignore_conflict:
                    return existing
                raise exceptions.UnicityError(id_field, existing)
            except exceptions.RecordNotFoundError:
                pass
        else:
            record[id_field] = id_generator()

        # Remove redundancy in data field
        query_record = {**record}
        query_record.pop(id_field, None)
        query_record.pop(modified_field, None)

        query = """
        WITH delete_potential_tombstone AS (
            DELETE FROM deleted
             WHERE id = :object_id
               AND parent_id = :parent_id
               AND collection_id = :collection_id
        )
        INSERT INTO records (id, parent_id, collection_id, data, last_modified)
        VALUES (:object_id, :parent_id,
                :collection_id, (:data)::JSONB,
                from_epoch(:last_modified))
        %(on_conflict)s
        RETURNING id, as_epoch(last_modified) AS last_modified;
        """

        safe_holders = {"on_conflict": ""}

        if ignore_conflict:
            # We use DO UPDATE so that the RETURNING clause works
            # but we don't update anything and keep the previous
            # last_modified value already stored.
            safe_holders["on_conflict"] = """
            ON CONFLICT (id, parent_id, collection_id) DO UPDATE
            SET last_modified = EXCLUDED.last_modified
            """

        placeholders = dict(object_id=record[id_field],
                            parent_id=parent_id,
                            collection_id=collection_id,
                            last_modified=record.get(modified_field),
                            data=json.dumps(query_record))
        with self.client.connect() as conn:
            result = conn.execute(query % safe_holders, placeholders)
            inserted = result.fetchone()

        record[modified_field] = inserted['last_modified']
        return record

    def get(self, collection_id, parent_id, object_id,
            id_field=DEFAULT_ID_FIELD,
            modified_field=DEFAULT_MODIFIED_FIELD,
            auth=None):
        query = """
        SELECT as_epoch(last_modified) AS last_modified, data
          FROM records
         WHERE id = :object_id
           AND parent_id = :parent_id
           AND collection_id = :collection_id;
        """
        placeholders = dict(object_id=object_id,
                            parent_id=parent_id,
                            collection_id=collection_id)
        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query, placeholders)
            if result.rowcount == 0:
                raise exceptions.RecordNotFoundError(object_id)
            else:
                existing = result.fetchone()

        record = existing['data']
        record[id_field] = object_id
        record[modified_field] = existing['last_modified']
        return record

    def update(self, collection_id, parent_id, object_id, record,
               id_field=DEFAULT_ID_FIELD,
               modified_field=DEFAULT_MODIFIED_FIELD,
               auth=None):

        # Remove redundancy in data field
        query_record = {**record}
        query_record.pop(id_field, None)
        query_record.pop(modified_field, None)

        query_create = """
        WITH delete_potential_tombstone AS (
            DELETE FROM deleted
             WHERE id = :object_id
               AND parent_id = :parent_id
               AND collection_id = :collection_id
        )
        INSERT INTO records (id, parent_id, collection_id, data, last_modified)
        VALUES (:object_id, :parent_id,
                :collection_id, (:data)::JSONB,
                from_epoch(:last_modified))
        RETURNING as_epoch(last_modified) AS last_modified;
        """

        query_update = """
        UPDATE records SET data=(:data)::JSONB,
                           last_modified=from_epoch(:last_modified)
        WHERE id = :object_id
           AND parent_id = :parent_id
           AND collection_id = :collection_id
        RETURNING as_epoch(last_modified) AS last_modified;
        """
        placeholders = dict(object_id=object_id,
                            parent_id=parent_id,
                            collection_id=collection_id,
                            last_modified=record.get(modified_field),
                            data=json.dumps(query_record))

        record = {**record, id_field: object_id}

        with self.client.connect() as conn:
            # Create or update ?
            query = """
            SELECT id FROM records
            WHERE id = :object_id
              AND parent_id = :parent_id
              AND collection_id = :collection_id;
            """
            result = conn.execute(query, placeholders)
            query = query_update if result.rowcount > 0 else query_create

            result = conn.execute(query, placeholders)
            updated = result.fetchone()

        record[modified_field] = updated['last_modified']
        return record

    def delete(self, collection_id, parent_id, object_id,
               id_field=DEFAULT_ID_FIELD, with_deleted=True,
               modified_field=DEFAULT_MODIFIED_FIELD,
               deleted_field=DEFAULT_DELETED_FIELD,
               auth=None, last_modified=None):
        if with_deleted:
            query = """
            WITH deleted_record AS (
                DELETE
                FROM records
                WHERE id = :object_id
                  AND parent_id = :parent_id
                  AND collection_id = :collection_id
                RETURNING id
            )
            INSERT INTO deleted (id, parent_id, collection_id, last_modified)
            SELECT id, :parent_id, :collection_id, from_epoch(:last_modified)
              FROM deleted_record
            RETURNING as_epoch(last_modified) AS last_modified;
            """
        else:
            query = """
                DELETE
                FROM records
                WHERE id = :object_id
                  AND parent_id = :parent_id
                  AND collection_id = :collection_id
                RETURNING as_epoch(last_modified) AS last_modified;
            """
        placeholders = dict(object_id=object_id,
                            parent_id=parent_id,
                            collection_id=collection_id,
                            last_modified=last_modified)

        with self.client.connect() as conn:
            result = conn.execute(query, placeholders)
            if result.rowcount == 0:
                raise exceptions.RecordNotFoundError(object_id)
            inserted = result.fetchone()

        record = {}
        record[modified_field] = inserted['last_modified']
        record[id_field] = object_id

        record[deleted_field] = True
        return record

    def delete_all(self, collection_id, parent_id, filters=None,
                   sorting=None, pagination_rules=None, limit=None,
                   id_field=DEFAULT_ID_FIELD, with_deleted=True,
                   modified_field=DEFAULT_MODIFIED_FIELD,
                   deleted_field=DEFAULT_DELETED_FIELD,
                   auth=None):
        if with_deleted:
            query = """
            WITH deleted_records AS (
                WITH matching_records AS (
                    SELECT id, parent_id, collection_id
                        FROM records
                        WHERE {parent_id_filter}
                              {collection_id_filter}
                              {conditions_filter}
                              {pagination_rules}
                        {sorting}
                        {pagination_limit}
                )
                DELETE
                FROM records
                USING matching_records
                WHERE records.id = matching_records.id
                  AND records.parent_id = matching_records.parent_id
                  AND records.collection_id = matching_records.collection_id
                RETURNING records.id, records.parent_id, records.collection_id
            )
            INSERT INTO deleted (id, parent_id, collection_id)
            SELECT id, parent_id, collection_id
              FROM deleted_records
            RETURNING id, as_epoch(last_modified) AS last_modified;
            """
        else:
            query = """
            WITH matching_records AS (
                SELECT id, parent_id, collection_id
                    FROM records
                    WHERE {parent_id_filter}
                          {collection_id_filter}
                          {conditions_filter}
                          {pagination_rules}
                    {sorting}
                    {pagination_limit}
            )
            DELETE
            FROM records
            USING matching_records
            WHERE records.id = matching_records.id
              AND records.parent_id = matching_records.parent_id
              AND records.collection_id = matching_records.collection_id
            RETURNING records.id, as_epoch(last_modified) AS last_modified;
            """

        id_field = id_field or self.id_field
        modified_field = modified_field or self.modified_field
        placeholders = dict(parent_id=parent_id,
                            collection_id=collection_id)
        # Safe strings
        safeholders = defaultdict(str)
        # Handle parent_id as a regex only if it contains *
        if '*' in parent_id:
            safeholders['parent_id_filter'] = 'parent_id LIKE :parent_id'
            placeholders['parent_id'] = parent_id.replace('*', '%')
        else:
            safeholders['parent_id_filter'] = 'parent_id = :parent_id'
        # If collection is None, remove it from query.
        if collection_id is None:
            safeholders['collection_id_filter'] = ''
        else:
            safeholders['collection_id_filter'] = 'AND collection_id = :collection_id'  # NOQA

        if filters:
            safe_sql, holders = self._format_conditions(filters,
                                                        id_field,
                                                        modified_field)
            safeholders['conditions_filter'] = 'AND {}'.format(safe_sql)
            placeholders.update(**holders)

        if sorting:
            sql, holders = self._format_sorting(sorting, id_field,
                                                modified_field)
            safeholders['sorting'] = sql
            placeholders.update(**holders)

        if pagination_rules:
            sql, holders = self._format_pagination(pagination_rules, id_field,
                                                   modified_field)
            safeholders['pagination_rules'] = 'AND {}'.format(sql)
            placeholders.update(**holders)

        if limit:
            # We validate the limit value in the resource class as integer.
            safeholders['pagination_limit'] = 'LIMIT {}'.format(limit)

        with self.client.connect() as conn:
            result = conn.execute(query.format_map(safeholders), placeholders)
            deleted = result.fetchmany(self._max_fetch_size)

        records = []
        for result in deleted:
            record = {}
            record[id_field] = result['id']
            record[modified_field] = result['last_modified']
            record[deleted_field] = True
            records.append(record)

        return records

    def purge_deleted(self, collection_id, parent_id, before=None,
                      id_field=DEFAULT_ID_FIELD,
                      modified_field=DEFAULT_MODIFIED_FIELD,
                      auth=None):
        query = """
        DELETE
        FROM deleted
        WHERE {parent_id_filter}
              {collection_id_filter}
              {conditions_filter};
        """
        id_field = id_field or self.id_field
        modified_field = modified_field or self.modified_field
        placeholders = dict(parent_id=parent_id,
                            collection_id=collection_id)
        # Safe strings
        safeholders = defaultdict(str)
        # Handle parent_id as a regex only if it contains *
        if '*' in parent_id:
            safeholders['parent_id_filter'] = 'parent_id LIKE :parent_id'
            placeholders['parent_id'] = parent_id.replace('*', '%')
        else:
            safeholders['parent_id_filter'] = 'parent_id = :parent_id'
        # If collection is None, remove it from query.
        if collection_id is None:
            safeholders['collection_id_filter'] = ''
        else:
            safeholders['collection_id_filter'] = 'AND collection_id = :collection_id'  # NOQA

        if before is not None:
            safeholders['conditions_filter'] = (
                'AND as_epoch(last_modified) < :before')
            placeholders['before'] = before

        with self.client.connect() as conn:
            result = conn.execute(query.format_map(safeholders), placeholders)

        return result.rowcount

    def get_all(self, collection_id, parent_id, filters=None, sorting=None,
                pagination_rules=None, limit=None, include_deleted=False,
                id_field=DEFAULT_ID_FIELD,
                modified_field=DEFAULT_MODIFIED_FIELD,
                deleted_field=DEFAULT_DELETED_FIELD,
                auth=None):
        query = """
        WITH total_filtered AS (
            SELECT COUNT(id) AS count
              FROM records
             WHERE {parent_id_filter}
               AND collection_id = :collection_id
               {conditions_filter}
        ),
        collection_filtered AS (
            SELECT id, last_modified, data
              FROM records
             WHERE {parent_id_filter}
               AND collection_id = :collection_id
               {conditions_filter}
             LIMIT {max_fetch_size}
        ),
        fake_deleted AS (
            SELECT (:deleted_field)::JSONB AS data
        ),
        filtered_deleted AS (
            SELECT id, last_modified, fake_deleted.data AS data
              FROM deleted, fake_deleted
             WHERE {parent_id_filter}
               AND collection_id = :collection_id
               {conditions_filter}
               {deleted_limit}
        ),
        all_records AS (
            SELECT * FROM filtered_deleted
             UNION ALL
            SELECT * FROM collection_filtered
        ),
        paginated_records AS (
            SELECT DISTINCT id
              FROM all_records
              {pagination_rules}
        )
        SELECT total_filtered.count AS count_total,
               a.id, as_epoch(a.last_modified) AS last_modified, a.data
          FROM paginated_records AS p JOIN all_records AS a ON (a.id = p.id),
               total_filtered
          {sorting}
          {pagination_limit};
        """
        deleted_field = json.dumps(dict([(deleted_field, True)]))

        # Unsafe strings escaped by PostgreSQL
        placeholders = dict(parent_id=parent_id,
                            collection_id=collection_id,
                            deleted_field=deleted_field)

        # Safe strings
        safeholders = defaultdict(str)
        safeholders['max_fetch_size'] = self._max_fetch_size

        # Handle parent_id as a regex only if it contains *
        if '*' in parent_id:
            safeholders['parent_id_filter'] = 'parent_id LIKE :parent_id'
            placeholders['parent_id'] = parent_id.replace('*', '%')
        else:
            safeholders['parent_id_filter'] = 'parent_id = :parent_id'

        if filters:
            safe_sql, holders = self._format_conditions(filters,
                                                        id_field,
                                                        modified_field)
            safeholders['conditions_filter'] = 'AND {}'.format(safe_sql)
            placeholders.update(**holders)

        if not include_deleted:
            safeholders['deleted_limit'] = 'LIMIT 0'

        if sorting:
            sql, holders = self._format_sorting(sorting, id_field,
                                                modified_field)
            safeholders['sorting'] = sql
            placeholders.update(**holders)

        if pagination_rules:
            sql, holders = self._format_pagination(pagination_rules, id_field,
                                                   modified_field)
            safeholders['pagination_rules'] = 'WHERE {}'.format(sql)
            placeholders.update(**holders)

        if limit:
            # We validate the limit value in the resource class as integer.
            safeholders['pagination_limit'] = 'LIMIT {}'.format(limit)

        with self.client.connect(readonly=True) as conn:
            result = conn.execute(query.format_map(safeholders), placeholders)
            retrieved = result.fetchmany(self._max_fetch_size)

        if not len(retrieved):
            return [], 0

        count_total = retrieved[0]['count_total']

        records = []
        for result in retrieved:
            record = result['data']
            record[id_field] = result['id']
            record[modified_field] = result['last_modified']
            records.append(record)

        return records, count_total

    def _format_conditions(self, filters, id_field, modified_field,
                           prefix='filters'):
        """Format the filters list in SQL, with placeholders for safe escaping.

        .. note::
            All conditions are combined using AND.

        .. note::

            Field name and value are escaped as they come from HTTP API.

        :returns: A SQL string with placeholders, and a dict mapping
            placeholders to actual values.
        :rtype: tuple
        """
        operators = {
            COMPARISON.EQ: '=',
            COMPARISON.NOT: '<>',
            COMPARISON.IN: 'IN',
            COMPARISON.EXCLUDE: 'NOT IN',
            COMPARISON.LIKE: 'ILIKE',
        }

        conditions = []
        holders = {}
        for i, filtr in enumerate(filters):
            value = filtr.value

            if filtr.field == id_field:
                sql_field = 'id'
                if isinstance(value, int):
                    value = str(value)
            elif filtr.field == modified_field:
                sql_field = 'as_epoch(last_modified)'
            else:
                column_name = "data"
                # Subfields: ``person.name`` becomes ``data->person->>name``
                subfields = filtr.field.split('.')
                for j, subfield in enumerate(subfields):
                    # Safely escape field name
                    field_holder = '{}_field_{}_{}'.format(prefix, i, j)
                    holders[field_holder] = subfield
                    # Use ->> to convert the last level to text.
                    column_name += "->>" if j == len(subfields) - 1 else "->"
                    column_name += ":{}".format(field_holder)

                # If field is missing, we default to ''.
                sql_field = "coalesce({}, '')".format(column_name)
                # Cast when comparing to number (eg. '4' < '12')
                try:
                    if value and (is_numeric(value) or all([is_numeric(v) for v in value])):
                        sql_field = "({})::numeric".format(column_name)
                except TypeError:  # not iterable
                    pass

            if filtr.operator not in (COMPARISON.IN, COMPARISON.EXCLUDE):
                # For the IN operator, let psycopg escape the values list.
                # Otherwise JSON-ify the native value (e.g. True -> 'true')
                if not isinstance(value, str):
                    value = json.dumps(value).strip('"')
            else:
                # If not all numeric, fallback to string comparison.
                if not all([is_numeric(v) for v in value]):
                    value = [str(v) for v in value]
                value = tuple(value)
                # WHERE field IN ();  -- Fails with syntax error.
                if len(value) == 0:
                    value = (None,)

            if filtr.operator == COMPARISON.LIKE:
                value = '%{}%'.format(value)

            # Safely escape value
            value_holder = '{}_value_{}'.format(prefix, i)
            holders[value_holder] = value

            sql_operator = operators.setdefault(filtr.operator,
                                                filtr.operator.value)
            cond = "{} {} :{}".format(sql_field, sql_operator, value_holder)
            conditions.append(cond)

        safe_sql = ' AND '.join(conditions)
        return safe_sql, holders

    def _format_pagination(self, pagination_rules, id_field, modified_field):
        """Format the pagination rules in SQL, with placeholders for
        safe escaping.

        .. note::

            All rules are combined using OR.

        .. note::

            Field names are escaped as they come from HTTP API.

        :returns: A SQL string with placeholders, and a dict mapping
            placeholders to actual values.
        :rtype: tuple
        """
        rules = []
        placeholders = {}

        for i, rule in enumerate(pagination_rules):
            prefix = 'rules_{}'.format(i)
            safe_sql, holders = self._format_conditions(rule,
                                                        id_field,
                                                        modified_field,
                                                        prefix=prefix)
            rules.append(safe_sql)
            placeholders.update(**holders)

        safe_sql = ' OR '.join(['({})'.format(r) for r in rules])
        return safe_sql, placeholders

    def _format_sorting(self, sorting, id_field, modified_field):
        """Format the sorting in SQL, with placeholders for safe escaping.

        .. note::

            Field names are escaped as they come from HTTP API.

        :returns: A SQL string with placeholders, and a dict mapping
            placeholders to actual values.
        :rtype: tuple
        """
        sorts = []
        holders = {}
        for i, sort in enumerate(sorting):

            if sort.field == id_field:
                sql_field = 'id'
            elif sort.field == modified_field:
                sql_field = 'last_modified'
            else:
                # Subfields: ``person.name`` becomes ``data->person->>name``
                subfields = sort.field.split('.')
                sql_field = 'data'
                for j, subfield in enumerate(subfields):
                    # Safely escape field name
                    field_holder = 'sort_field_{}_{}'.format(i, j)
                    holders[field_holder] = subfield
                    # Use ->> to convert the last level to text.
                    sql_field += '->(:{})'.format(field_holder)

            sql_direction = 'ASC' if sort.direction > 0 else 'DESC'
            sql_sort = "{} {}".format(sql_field, sql_direction)
            sorts.append(sql_sort)

        safe_sql = 'ORDER BY {}'.format(', '.join(sorts))
        return safe_sql, holders


def load_from_config(config):
    settings = config.get_settings()
    max_fetch_size = int(settings['storage_max_fetch_size'])
    client = create_from_config(config, prefix='storage_')
    return Storage(client=client, max_fetch_size=max_fetch_size)



import unittest
import mock

from pyramid import httpexceptions

from kinto.core.errors import ERRORS, http_error
from kinto.core.testing import FormattedErrorMixin
from kinto.core.storage import exceptions as storage_exceptions

from .support import BaseWebTest, authorize


class ErrorViewTest(FormattedErrorMixin, BaseWebTest, unittest.TestCase):

    sample_url = "/mushrooms"

    def test_backoff_headers_is_not_present_by_default(self):
        response = self.app.get(self.sample_url, headers=self.headers, status=200)
        self.assertNotIn('Backoff', response.headers)

    def test_backoff_headers_is_present_if_configured(self):
        with mock.patch.dict(self.app.app.registry.settings, [('backoff', 10)]):
            response = self.app.get(self.sample_url, headers=self.headers, status=200)
        self.assertIn('Backoff', response.headers)

    def test_backoff_headers_is_present_on_304(self):
        first = self.app.get(self.sample_url, headers=self.headers)
        etag = first.headers['ETag']
        headers = {**self.headers, 'If-None-Match': etag}
        with mock.patch.dict(self.app.app.registry.settings, [('backoff', 10)]):
            response = self.app.get(self.sample_url, headers=headers, status=304)
        self.assertIn('Backoff', response.headers)

    def test_backoff_header_is_present_on_error_responses(self):
        with mock.patch.dict(
                self.app.app.registry.settings,
                [('backoff', 10)]):
            response = self.app.get(self.sample_url, headers=self.headers, status=200)
            self.assertIn('Backoff', response.headers)
            self.assertEquals(response.headers['Backoff'], '10')

    def test_404_is_valid_formatted_error(self):
        response = self.app.get('/unknown', status=404)
        self.assertFormattedError(response, 404, ERRORS.MISSING_RESOURCE, "Not Found",
                                  "The resource you are looking for could not be found.")

    def test_404_can_be_overridden(self):
        custom_404 = http_error(httpexceptions.HTTPNotFound(),
                                errno=ERRORS.MISSING_RESOURCE,
                                message="Customized.")
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=custom_404):
            response = self.app.get(self.sample_url, headers=self.headers, status=404)
        self.assertFormattedError(response, 404, ERRORS.MISSING_RESOURCE,
                                  "Not Found", "Customized.")

    def test_401_is_valid_formatted_error(self):
        response = self.app.get(self.sample_url, status=401)
        self.assertFormattedError(response, 401, ERRORS.MISSING_AUTH_TOKEN, "Unauthorized",
                                  "Please authenticate yourself to use this endpoint.")

    @authorize(False)
    def test_403_is_valid_formatted_error(self):
        response = self.app.get(self.sample_url, headers=self.headers, status=403)
        self.assertFormattedError(response, 403, ERRORS.FORBIDDEN, "Forbidden",
                                  "This user cannot access this resource.")

    def test_403_can_be_overridded(self):
        custom_403 = http_error(httpexceptions.HTTPForbidden(), errno=ERRORS.FORBIDDEN,
                                message="Customized.")
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=custom_403):
            response = self.app.get(self.sample_url, headers=self.headers, status=403)
        self.assertFormattedError(response, 403, ERRORS.FORBIDDEN, "Forbidden", "Customized.")

    def test_405_is_valid_formatted_error(self):
        response = self.app.patch(self.sample_url, headers=self.headers, status=405)
        self.assertFormattedError(response, 405, ERRORS.METHOD_NOT_ALLOWED, "Method Not Allowed",
                                  "Method not allowed on this endpoint.")

    def test_405_can_have_custom_message(self):
        custom_405 = http_error(httpexceptions.HTTPMethodNotAllowed(),
                                errno=ERRORS.METHOD_NOT_ALLOWED, message="Disabled from conf.")
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=custom_405):
            response = self.app.get(self.sample_url, headers=self.headers, status=405)
        self.assertFormattedError(response, 405, ERRORS.METHOD_NOT_ALLOWED, "Method Not Allowed",
                                  "Disabled from conf.")

    def test_500_is_valid_formatted_error(self):
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=ValueError):
            response = self.app.get(self.sample_url, headers=self.headers, status=500)
        self.assertFormattedError(
            response, 500, ERRORS.UNDEFINED, "Internal Server Error",
            "A programmatic error occured, developers have been informed.",
            "https://github.com/Kinto/kinto/issues/")

    def test_400_with_invalid_url_path(self):
        response = self.app.get('/%82%AC', status=400)
        self.assertFormattedError(response, 400, ERRORS.INVALID_PARAMETERS, "Bad Request",
                                  "Invalid URL path.")

    def test_info_link_in_error_responses_can_be_configured(self):
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=ValueError):
            link = "https://github.com/mozilla-services/kinto/issues/"
            app = self.make_app({'error_info_link': link, 'readonly': False})
            response = app.get(self.sample_url, headers=self.headers, status=500)
        self.assertFormattedError(response, 500, ERRORS.UNDEFINED, "Internal Server Error",
                                  "A programmatic error occured, developers have been informed.",
                                  link)

    def test_503_is_valid_formatted_error(self):
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=httpexceptions.HTTPServiceUnavailable):
            response = self.app.get(self.sample_url, headers=self.headers, status=503)
        self.assertFormattedError(response, 503, ERRORS.BACKEND, "Service Unavailable",
                                  ("Service temporary unavailable "
                                   "due to overloading or maintenance, please retry later."))
        self.assertIn("Retry-After", response.headers)

    def test_503_can_have_custom_message(self):
        custom_503 = http_error(httpexceptions.HTTPServiceUnavailable(),
                                errno=ERRORS.BACKEND,
                                message="Unable to connect the server.")
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=custom_503):
            response = self.app.get(self.sample_url, headers=self.headers, status=503)
        self.assertFormattedError(response, 503, ERRORS.BACKEND, "Service Unavailable",
                                  "Unable to connect the server.")

    def test_integrity_errors_are_served_as_409(self):
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=storage_exceptions.IntegrityError):
            response = self.app.get(self.sample_url, headers=self.headers, status=409)
        self.assertFormattedError(response, 409, ERRORS.CONSTRAINT_VIOLATED, "Conflict",
                                  "Integrity constraint violated, please retry.")
        self.assertIn("Retry-After", response.headers)

    def test_500_provides_traceback_on_server(self):
        mock_traceback = mock.patch('logging.traceback.print_exception')
        with mock.patch('tests.core.testapp.views.Mushroom._extract_filters',
                        side_effect=ValueError):
            with mock_traceback as mocked_traceback:
                self.app.get(self.sample_url, headers=self.headers, status=500)
                self.assertTrue(mocked_traceback.called)
                self.assertEqual(ValueError, mocked_traceback.call_args[0][0])


class RedirectViewTest(FormattedErrorMixin, BaseWebTest, unittest.TestCase):
    api_prefix = ''

    def test_do_not_redirect_to_version_if_not_supported_version(self):
        resp = self.app.get('/v1/', status=404)
        self.assertFormattedError(resp, 404, ERRORS.VERSION_NOT_AVAILABLE, "Not Found",
                                  "The requested API version is not available on this server.")

    def test_redirect_to_version(self):
        # GET on the hello view.
        response = self.app.get('/')
        self.assertEqual(response.status_int, 307)
        self.assertEqual(response.location, 'http://localhost/v0/')

        # GET on the fields view.
        response = self.app.get('/mushrooms')
        self.assertEqual(response.status_int, 307)
        self.assertEqual(response.location, 'http://localhost/v0/mushrooms')

    def test_redirects_benefits_from_cors_setup(self):
        headers = {
            'Origin': 'lolnet.org',
            'Access-Control-Request-Method': 'GET'
        }
        resp = self.app.options('/', headers=headers, status=200)
        self.assertIn('Access-Control-Allow-Origin', resp.headers)

    def test_do_not_redirect_to_version_if_disabled_in_settings(self):
        # GET on the hello view.
        app = self.make_app({'version_prefix_redirect_enabled': False})
        app.get('/', status=404)

    def test_querystring_is_preserved_during_redirection(self):
        response = self.app.get('/home/articles?_since=42')
        self.assertEqual(response.location, 'http://localhost/v0/home/articles?_since=42')

    def test_redirection_does_not_allow_crlf(self):
        self.app.get('/crlftest%0DSet-Cookie:test%3Dtest%3Bdomain%3D.yelp.com',
                     status=404)

    def test_redirection_does_not_allow_control_characters(self):
        self.app.get('/9l2j7%0A21m2n', status=404)

    def test_redirection_allows_unicode_characters(self):
        # URL with unicode: /crlftest嘊/
        self.app.get('/crlftest%E5%98%8A/', status=307)

    def test_redirection_allows_unicode_characters_in_querystring(self):
        # URL with unicode: /crlftest?name=嘊
        self.app.get('/crlftest?name=%E5%98%8A', status=307)


class TrailingSlashRedirectViewTest(FormattedErrorMixin, BaseWebTest,
                                    unittest.TestCase):
    def test_doesnt_redirect_the_home_page(self):
        response = self.app.get('/')
        self.assertEqual(response.status_int, 200)

    def test_does_redirect_the_version_prefix(self):
        response = self.app.get('')
        self.assertEqual(response.status_int, 307)
        self.assertEqual(response.location, 'http://localhost/v0/')

    def test_it_redirects_if_it_ends_with_a__slash_(self):
        response = self.app.get('/mushrooms/')
        self.assertEqual(response.status_int, 307)
        self.assertEqual(response.location, 'http://localhost/v0/mushrooms')

    def test_do_not_redirect_if_disabled_in_settings(self):
        app = self.make_app({'trailing_slash_redirect_enabled': False})
        app.get('/', status=200)
        app.get('/mushrooms/', status=404)

    def test_querystring_is_preserved_during_redirection(self):
        response = self.app.get('/home/articles/?_since=42')
        self.assertEqual(response.location, 'http://localhost/v0/home/articles?_since=42')

    def test_it_does_not_redirect_if_the_url_exists(self):
        self.app.get('/static/', status=200)

    def test_display_an_error_message_if_disabled_in_settings(self):
        app = self.make_app({'trailing_slash_redirect_enabled': False})
        response = app.get('', status=404)
        self.assertFormattedError(response, 404, ERRORS.MISSING_RESOURCE, "Not Found",
                                  "The resource you are looking for could not be found.")



from pyramid import httpexceptions

from . import BaseTest


class ModelTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.record = self.model.create_record({'field': 'value'})

    def test_list_returns_all_records_in_data(self):
        result = self.resource.collection_get()
        records = result['data']
        self.assertEqual(len(records), 1)
        self.assertDictEqual(records[0], self.record)


class CreateTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.resource.request.validated['body'] = {'data': {'field': 'new'}}

    def test_new_records_are_linked_to_owner(self):
        resp = self.resource.collection_post()['data']
        record_id = resp['id']
        self.model.get_record(record_id)  # not raising

    def test_create_record_returns_at_least_id_and_last_modified(self):
        record = self.resource.collection_post()['data']
        self.assertIn(self.resource.model.id_field, record)
        self.assertIn(self.resource.model.modified_field, record)
        self.assertIn('field', record)


class DeleteModelTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.patch_known_field.start()
        self.model.create_record({'field': 'a'})
        self.model.create_record({'field': 'b'})

    def test_delete_on_list_removes_all_records(self):
        self.resource.collection_delete()
        result = self.resource.collection_get()
        records = result['data']
        self.assertEqual(len(records), 0)

    def test_delete_returns_deleted_version_of_records(self):
        result = self.resource.collection_delete()
        deleted = result['data'][0]
        self.assertIn('deleted', deleted)

    def test_delete_supports_collection_filters(self):
        self.resource.request.validated['querystring'] = {'field': 'a'}
        self.resource.collection_delete()
        self.resource.request.validated['querystring'] = {}
        result = self.resource.collection_get()
        records = result['data']
        self.assertEqual(len(records), 1)


class IsolatedModelsTest(BaseTest):
    def setUp(self):
        super().setUp()
        self.resource.request.validated = {'header': {}, 'querystring': {}}
        self.stored = self.model.create_record({}, parent_id='bob')
        self.resource.record_id = self.stored['id']

    def get_request(self):
        request = super().get_request()
        request.prefixed_userid = 'basicauth:alice'
        return request

    def get_context(self):
        context = super().get_context()
        context.prefixed_userid = 'basicauth:alice'
        return context

    def test_list_is_filtered_by_user(self):
        resp = self.resource.collection_get()
        records = resp['data']
        self.assertEquals(len(records), 0)

    def test_update_record_of_another_user_will_create_it(self):
        self.resource.request.validated['body'] = {'data': {'some': 'record'}}
        self.resource.put()
        self.model.get_record(record_id=self.stored['id'],
                              parent_id='basicauth:alice')  # not raising

    def test_cannot_modify_record_of_other_user(self):
        self.assertRaises(httpexceptions.HTTPNotFound, self.resource.patch)

    def test_cannot_delete_record_of_other_user(self):
        self.assertRaises(httpexceptions.HTTPNotFound, self.resource.delete)



from collections import OrderedDict

import transaction
from pyramid.events import NewRequest
from enum import Enum

from kinto.core.logs import logger
from kinto.core.utils import strip_uri_prefix


class ACTIONS(Enum):
    CREATE = 'create'
    DELETE = 'delete'
    READ = 'read'
    UPDATE = 'update'

    @staticmethod
    def from_string_list(elements):
        return tuple(ACTIONS(el) for el in elements)


class _ResourceEvent:
    def __init__(self, payload, request):
        self.payload = payload
        self.request = request

    def __repr__(self):
        return "<{klass} action={action} uri={uri}>".format(
            klass=self.__class__.__name__,
            **self.payload)


class ResourceRead(_ResourceEvent):
    """Triggered when a resource is being read.
    """
    def __init__(self, payload, read_records, request):
        super().__init__(payload, request)
        self.read_records = read_records


class ResourceChanged(_ResourceEvent):
    """Triggered when a resource is being changed.
    """
    def __init__(self, payload, impacted_records, request):
        super().__init__(payload, request)
        self.impacted_records = impacted_records


class AfterResourceRead(_ResourceEvent):
    """Triggered after a resource was successfully read.
    """
    def __init__(self, payload, read_records, request):
        super().__init__(payload, request)
        self.read_records = read_records


class AfterResourceChanged(_ResourceEvent):
    """Triggered after a resource was successfully changed.
    """
    def __init__(self, payload, impacted_records, request):
        super().__init__(payload, request)
        self.impacted_records = impacted_records


def setup_transaction_hook(config):
    """
    Resource events are plugged with the transactions of ``pyramid_tm``.

    Once a transaction is committed, ``AfterResourceRead`` and
    ``AfterResourceChanged`` events are sent.
    """
    def _notify_resource_events_before(request):
        """Notify the accumulated resource events before end of transaction.
        """
        for event in request.get_resource_events():
            request.registry.notify(event)

    def _notify_resource_events_after(success, request):
        """Notify the accumulated resource events if transaction succeeds.
        """
        if not success:  # pragma: no cover
            return

        for event in request.get_resource_events(after_commit=True):
            try:
                request.registry.notify(event)
            except Exception:
                logger.error("Unable to notify", exc_info=True)

    def on_new_request(event):
        """When a new request comes in, hook on transaction commit.
        """
        # Since there is one transaction per batch, ignore subrequests.
        if hasattr(event.request, 'parent'):
            return
        current = transaction.get()
        current.addBeforeCommitHook(_notify_resource_events_before,
                                    args=(event.request,))
        current.addAfterCommitHook(_notify_resource_events_after,
                                   args=(event.request,))

    config.add_subscriber(on_new_request, NewRequest)


def get_resource_events(request, after_commit=False):
    """
    Request helper to return the list of events triggered on resources.
    The list is sorted chronologically (see OrderedDict)
    """
    by_resource = request.bound_data.get("resource_events", {})
    events = []
    for (action, payload, impacted, request) in by_resource.values():
        if after_commit:
            if action == ACTIONS.READ:
                event_cls = AfterResourceRead
            else:
                event_cls = AfterResourceChanged
        else:
            if action == ACTIONS.READ:
                event_cls = ResourceRead
            else:
                event_cls = ResourceChanged
        event = event_cls(payload, impacted, request)
        events.append(event)
    return events


def notify_resource_event(request, parent_id, timestamp, data, action,
                          old=None):
    """
    Request helper to stack a resource event.

    If a similar event (same resource, same action) already occured during the
    current transaction (e.g. batch) then just extend the impacted records of
    the previous one.
    """
    if action == ACTIONS.READ:
        if not isinstance(data, list):
            data = [data]
        impacted = data
    elif action == ACTIONS.CREATE:
        impacted = [{'new': data}]
    elif action == ACTIONS.DELETE:
        if not isinstance(data, list):
            impacted = [{'new': data, 'old': old}]
        else:
            impacted = []
            for i, new in enumerate(data):
                impacted.append({'new': new, 'old': old[i]})
    else:  # ACTIONS.UPDATE:
        impacted = [{'new': data, 'old': old}]

    # Get previously triggered events.
    events = request.bound_data.setdefault("resource_events", OrderedDict())

    resource_name = request.current_resource_name

    # Group events by resource and action.
    group_by = resource_name + parent_id + action.value

    if group_by in events:
        # Add to impacted records of existing event.
        already_impacted = events[group_by][2]
        already_impacted.extend(impacted)
    else:
        # Create new event.
        payload = {'timestamp': timestamp,
                   'action': action.value,
                   'uri': strip_uri_prefix(request.path),
                   'user_id': request.prefixed_userid,
                   'resource_name': resource_name}

        matchdict = dict(request.matchdict)

        if 'id' in request.matchdict:
            matchdict[resource_name + '_id'] = matchdict.pop('id')

        payload.update(**matchdict)

        events[group_by] = (action, payload, impacted, request)



"""This module contains generic schemas used by the core. You should declare schemas that
may be reused across the `kinto.core` here.

.. note:: This module is reserve for generic schemas only.

    - If a schema is resource specific, you should use `kinto.core.resource.schema`.
    - If a schema is view specific, you should declare it on the respective view.
"""

import colander

from kinto.core.utils import strip_whitespace, msec_time, native_value


class TimeStamp(colander.SchemaNode):
    """Basic integer schema field that can be set to current server timestamp
    in milliseconds if no value is provided.

    .. code-block:: python

        class Book(ResourceSchema):
            added_on = TimeStamp()
            read_on = TimeStamp(auto_now=False, missing=-1)
    """
    schema_type = colander.Integer

    title = 'Epoch timestamp'
    """Default field title."""

    auto_now = True
    """Set to current server timestamp (*milliseconds*) if not provided."""

    missing = None
    """Default field value if not provided in record."""

    def deserialize(self, cstruct=colander.null):
        if cstruct is colander.null and self.auto_now:
            cstruct = msec_time()
        return super(TimeStamp, self).deserialize(cstruct)


class URL(colander.SchemaNode):
    """String field representing a URL, with max length of 2048.
    This is basically a shortcut for string field with
    `~colander:colander.url`.

    .. code-block:: python

        class BookmarkSchema(ResourceSchema):
            url = URL()
    """
    schema_type = colander.String
    validator = colander.All(colander.url, colander.Length(min=1, max=2048))

    def preparer(self, appstruct):
        return strip_whitespace(appstruct)


class Any(colander.SchemaType):
    """Colander type agnostic field."""

    def deserialize(self, node, cstruct):
        return cstruct


class HeaderField(colander.SchemaNode):
    """Basic header field SchemaNode."""

    missing = colander.drop

    def deserialize(self, cstruct=colander.null):
        if isinstance(cstruct, bytes):
            try:
                cstruct = cstruct.decode('utf-8')
            except UnicodeDecodeError:
                raise colander.Invalid(self, msg='Headers should be UTF-8 encoded')
        return super(HeaderField, self).deserialize(cstruct)


class QueryField(colander.SchemaNode):
    """Basic querystring field SchemaNode."""

    missing = colander.drop

    def deserialize(self, cstruct=colander.null):
        if isinstance(cstruct, str):
            cstruct = native_value(cstruct)
        return super(QueryField, self).deserialize(cstruct)


class FieldList(QueryField):
    """String field representing a list of attributes."""

    schema_type = colander.Sequence
    error_message = "The value should be a list of comma separated attributes"
    missing = colander.drop
    fields = colander.SchemaNode(colander.String(), missing=colander.drop)

    def deserialize(self, cstruct=colander.null):
        if isinstance(cstruct, str):
            cstruct = cstruct.split(',')
        return super(FieldList, self).deserialize(cstruct)


class HeaderQuotedInteger(HeaderField):
    """Integer between "" used in precondition headers."""

    schema_type = colander.String
    error_message = "The value should be integer between double quotes."
    validator = colander.Regex('^"([0-9]+?)"$|\*', msg=error_message)

    def deserialize(self, cstruct=colander.null):
        param = super(HeaderQuotedInteger, self).deserialize(cstruct)
        if param is colander.drop or param == '*':
            return param

        return int(param[1:-1])



"""Main entry point
"""
import pkg_resources
import tempfile

from cornice import Service as CorniceService
from pyramid.settings import aslist

from kinto.core import errors
from kinto.core import events
from kinto.core.initialization import (  # NOQA
    initialize, install_middlewares,
    load_default_settings)
from kinto.core.utils import (
    follow_subrequest, current_service, current_resource_name,
    prefixed_userid, prefixed_principals)
from kinto.core.logs import logger


# Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution('kinto').version  # FIXME?

DEFAULT_SETTINGS = {
    'backoff': None,
    'batch_max_requests': 25,
    'cache_backend': '',
    'cache_url': '',
    'cache_pool_size': 25,
    'cache_prefix': '',
    'cache_max_size_bytes': 524288,
    'cors_origins': '*',
    'cors_max_age_seconds': 3600,
    'eos': None,
    'eos_message': None,
    'eos_url': None,
    'error_info_link': 'https://github.com/Kinto/kinto/issues/',
    'http_host': None,
    'http_scheme': None,
    'id_generator': 'kinto.core.storage.generators.UUID4',
    'includes': '',
    'initialization_sequence': (
        'kinto.core.initialization.setup_request_bound_data',
        'kinto.core.initialization.setup_json_serializer',
        'kinto.core.initialization.setup_logging',
        'kinto.core.initialization.setup_storage',
        'kinto.core.initialization.setup_permission',
        'kinto.core.initialization.setup_cache',
        'kinto.core.initialization.setup_requests_scheme',
        'kinto.core.initialization.setup_version_redirection',
        'kinto.core.initialization.setup_deprecation',
        'kinto.core.initialization.setup_authentication',
        'kinto.core.initialization.setup_backoff',
        'kinto.core.initialization.setup_statsd',
        'kinto.core.initialization.setup_listeners',
        'kinto.core.events.setup_transaction_hook',
    ),
    'event_listeners': '',
    'heartbeat_timeout_seconds': 10,
    'logging_renderer': 'kinto.core.logs.ClassicLogRenderer',
    'newrelic_config': None,
    'newrelic_env': 'dev',
    'paginate_by': None,
    'permission_backend': '',
    'permission_url': '',
    'permission_pool_size': 25,
    'profiler_dir': tempfile.gettempdir(),
    'profiler_enabled': False,
    'project_docs': '',
    'project_name': '',
    'project_version': '',
    'readonly': False,
    'retry_after_seconds': 30,
    'statsd_backend': 'kinto.core.statsd',
    'statsd_prefix': 'kinto.core',
    'statsd_url': None,
    'storage_backend': '',
    'storage_url': '',
    'storage_max_fetch_size': 10000,
    'storage_pool_size': 25,
    'tm.annotate_user': False,  # Do annotate transactions with the user-id.
    'transaction_per_request': True,
    'userid_hmac_secret': '',
    'version_json_path': 'version.json',
    'version_prefix_redirect_enabled': True,
    'trailing_slash_redirect_enabled': True,
    'multiauth.groupfinder': 'kinto.core.authorization.groupfinder',
    'multiauth.policies': 'basicauth',
    'multiauth.policy.basicauth.use': ('kinto.core.authentication.'
                                       'BasicAuthAuthenticationPolicy'),
    'multiauth.authorization_policy': ('kinto.core.authorization.'
                                       'AuthorizationPolicy'),
    'swagger_file': 'swagger.yaml',
}


class Service(CorniceService):
    """Subclass of the default cornice service.

    This is useful in order to attach specific behaviours without monkey
    patching the default cornice service (which would impact other uses of it)
    """
    default_cors_headers = ('Backoff', 'Retry-After', 'Alert',
                            'Content-Length')

    def error_handler(self, request):
        return errors.json_error_handler(request)

    @classmethod
    def init_from_settings(cls, settings):
        cls.cors_origins = tuple(aslist(settings['cors_origins']))
        cors_max_age = settings['cors_max_age_seconds']
        cls.cors_max_age = int(cors_max_age) if cors_max_age else None


def includeme(config):
    settings = config.get_settings()

    # Heartbeat registry.
    config.registry.heartbeats = {}

    # Public settings registry.
    config.registry.public_settings = {'batch_max_requests', 'readonly'}

    # Directive to declare arbitrary API capabilities.
    def add_api_capability(config, identifier, description="", url="", **kw):
        existing = config.registry.api_capabilities.get(identifier)
        if existing:
            error_msg = "The '{}' API capability was already registered ({})."
            raise ValueError(error_msg.format(identifier, existing))

        capability = dict(description=description, url=url, **kw)
        config.registry.api_capabilities[identifier] = capability

    config.add_directive('add_api_capability', add_api_capability)
    config.registry.api_capabilities = {}

    # Resource events helpers.
    config.add_request_method(events.get_resource_events,
                              name='get_resource_events')
    config.add_request_method(events.notify_resource_event,
                              name='notify_resource_event')

    # Setup cornice.
    config.include("cornice")

    # Setup cornice api documentation
    config.include("cornice_swagger")

    # Per-request transaction.
    config.include("pyramid_tm")

    # Add CORS settings to the base kinto.core Service class.
    Service.init_from_settings(settings)

    # Setup components.
    for step in aslist(settings['initialization_sequence']):
        step_func = config.maybe_dotted(step)
        step_func(config)

    # Custom helpers.
    config.add_request_method(follow_subrequest)
    config.add_request_method(prefixed_userid, property=True)
    config.add_request_method(prefixed_principals, reify=True)
    config.add_request_method(lambda r: {
        'id': r.prefixed_userid,
        'principals': r.prefixed_principals},
        name='get_user_info')
    config.add_request_method(current_resource_name, reify=True)
    config.add_request_method(current_service, reify=True)
    config.commit()

    # Include plugins after init, unlike pyramid includes.
    includes = aslist(settings['includes'])
    for app in includes:
        config.include(app)

    # # Show settings to output.
    # for key, value in settings.items():
    #     logger.info('Using {} = {}'.format(key, value))

    # Scan views.
    config.scan("kinto.core.views")

    # Give sign of life.
    msg = "Running {project_name} {project_version}."
    logger.info(msg.format_map(settings))



import mock
import time

from pyramid import testing

from kinto.core.storage import exceptions
from kinto.core.cache import heartbeat


class CacheTest:
    backend = None
    settings = {}

    def setUp(self):
        super().setUp()
        self.cache = self.backend.load_from_config(self._get_config())
        self.cache.initialize_schema()
        self.request = None
        self.client_error_patcher = None

    def _get_config(self, settings=None):
        """Mock Pyramid config object.
        """
        if settings is None:
            settings = self.settings
        config = testing.setUp()
        config.add_settings(settings)
        return config

    def tearDown(self):
        mock.patch.stopall()
        super().tearDown()
        self.cache.flush()

    def get_backend_prefix(self, prefix):
        settings_prefix = {**self.settings}
        settings_prefix['cache_prefix'] = prefix
        config_prefix = self._get_config(settings=settings_prefix)

        # initiating cache backend with prefix:
        backend_prefix = self.backend.load_from_config(config_prefix)

        return backend_prefix

    def test_backend_error_is_raised_anywhere(self):
        self.client_error_patcher.start()
        calls = [
            (self.cache.flush,),
            (self.cache.ttl, ''),
            (self.cache.expire, '', 0),
            (self.cache.get, ''),
            (self.cache.set, '', ''),
            (self.cache.delete, ''),
        ]
        for call in calls:
            self.assertRaises(exceptions.BackendError, *call)

    def test_initialize_schema_is_idempotent(self):
        self.cache.initialize_schema()
        self.cache.initialize_schema()  # not raising.

    def test_ping_returns_false_if_unavailable(self):
        self.client_error_patcher.start()
        ping = heartbeat(self.cache)
        self.assertFalse(ping(self.request))
        with mock.patch('kinto.core.cache.random.SystemRandom.random', return_value=0.6):
            self.assertFalse(ping(self.request))
        with mock.patch('kinto.core.cache.random.SystemRandom.random', return_value=0.4):
            self.assertFalse(ping(self.request))

    def test_ping_returns_true_if_available(self):
        ping = heartbeat(self.cache)
        with mock.patch('kinto.core.cache.random.random', return_value=0.6):
            self.assertTrue(ping(self.request))
        with mock.patch('kinto.core.cache.random.random', return_value=0.4):
            self.assertTrue(ping(self.request))

    def test_ping_logs_error_if_unavailable(self):
        self.client_error_patcher.start()
        ping = heartbeat(self.cache)

        with mock.patch('kinto.core.cache.logger.exception') as exc_handler:
            self.assertFalse(ping(self.request))

        self.assertTrue(exc_handler.called)

    def test_set_adds_the_record(self):
        stored = 'toto'
        self.cache.set('foobar', stored)
        retrieved = self.cache.get('foobar')
        self.assertEquals(retrieved, stored)

    def test_values_remains_python_dict(self):
        def setget(k, v):
            self.cache.set(k, v)
            return (self.cache.get(k), v)

        self.assertEqual(*setget('foobar', 3))
        self.assertEqual(*setget('foobar', ['a']))
        self.assertEqual(*setget('foobar', {'b': [1, 2]}))
        self.assertEqual(*setget('foobar', 3.14))

    def test_delete_removes_the_record(self):
        self.cache.set('foobar', 'toto')
        self.cache.delete('foobar')
        retrieved = self.cache.get('foobar')
        self.assertIsNone(retrieved)

    def test_delete_does_not_fail_if_record_is_unknown(self):
        self.cache.delete('foobar')

    def test_expire_expires_the_value(self):
        self.cache.set('foobar', 'toto')
        self.cache.expire('foobar', 0.01)
        time.sleep(0.02)
        retrieved = self.cache.get('foobar')
        self.assertIsNone(retrieved)

    def test_set_with_ttl_expires_the_value(self):
        self.cache.set('foobar', 'toto', 0.01)
        time.sleep(0.02)
        retrieved = self.cache.get('foobar')
        self.assertIsNone(retrieved)

    def test_ttl_return_the_time_to_live(self):
        self.cache.set('foobar', 'toto')
        self.cache.expire('foobar', 10)
        ttl = self.cache.ttl('foobar')
        self.assertGreater(ttl, 0)
        self.assertLessEqual(ttl, 10)

    def test_ttl_return_none_if_unknown(self):
        ttl = self.cache.ttl('unknown')
        self.assertTrue(ttl < 0)

    def test_cache_prefix_is_set(self):
        backend_prefix = self.get_backend_prefix(prefix='prefix_')

        # Set the value
        backend_prefix.set('key', 'foo')

        # Validate that it was set with the prefix.
        obtained = self.cache.get('prefix_key')
        self.assertEqual(obtained, 'foo')

    def test_cache_when_prefix_is_not_set(self):
        backend_prefix = self.get_backend_prefix(prefix='')

        # Set a value
        backend_prefix.set('key', 'foo')

        # Validate that it was set with no prefix
        obtained = self.cache.get('key')
        self.assertEqual(obtained, 'foo')

    def test_prefix_value_use_to_get_data(self):
        backend_prefix = self.get_backend_prefix(prefix='prefix_')

        # Set the value with the prefix
        self.cache.set('prefix_key', 'foo')

        # Validate that the prefix was added
        obtained = backend_prefix.get('key')
        self.assertEqual(obtained, 'foo')

    def test_prefix_value_use_to_delete_data(self):
        backend_prefix = self.get_backend_prefix(prefix='prefix_')
        # Set the value
        self.cache.set('prefix_key', 'foo')

        # Delete the value
        backend_prefix.delete('key')

        # Validate that the value was deleted
        obtained = self.cache.get('prefix_key')
        self.assertEqual(obtained, None)

    def test_prefix_value_used_with_ttl(self):
        backend_prefix = self.get_backend_prefix(prefix='prefix_')

        self.cache.set('prefix_key', 'foo', 10)

        # Validate that the ttl add the prefix to the key.
        obtained = backend_prefix.ttl('key')
        self.assertLessEqual(obtained, 10)
        self.assertGreater(obtained, 9)

    def test_prefix_value_used_with_expire(self):
        backend_prefix = self.get_backend_prefix(prefix='prefix_')

        self.cache.set('prefix_foobar', 'toto', 10)

        # expiring the ttl of key
        backend_prefix.expire('foobar', 0)

        # Make sure the TTL was set accordingly.
        ttl = self.cache.ttl('prefix_foobar')
        self.assertLessEqual(ttl, 0)

        # The record should have expired
        retrieved = self.cache.get('prefix_foobar')
        self.assertIsNone(retrieved)



import types

try:
    import statsd as statsd_module
except ImportError:  # pragma: no cover
    statsd_module = None

from pyramid.exceptions import ConfigurationError
from urllib.parse import urlparse

from kinto.core import utils


class Client:
    def __init__(self, host, port, prefix):
        self._client = statsd_module.StatsClient(host, port, prefix=prefix)

    def watch_execution_time(self, obj, prefix='', classname=None):
        classname = classname or utils.classname(obj)
        members = dir(obj)
        for name in members:
            value = getattr(obj, name)
            is_method = isinstance(value, types.MethodType)
            if not name.startswith('_') and is_method:
                statsd_key = "{}.{}.{}".format(prefix, classname, name)
                decorated_method = self.timer(statsd_key)(value)
                setattr(obj, name, decorated_method)

    def timer(self, key):
        return self._client.timer(key)

    def count(self, key, unique=None):
        if unique is None:
            return self._client.incr(key, count=1)
        else:
            return self._client.set(key, unique)


def statsd_count(request, count_key):
    statsd = request.registry.statsd
    if statsd:
        statsd.count(count_key)


def load_from_config(config):
    # If this is called, it means that a ``statsd_url`` was specified in settings.
    # (see ``kinto.core.initialization``)
    # Raise a proper error if the ``statsd`` module is not installed.
    if statsd_module is None:
        error_msg = "Please install Kinto with monitoring dependencies (e.g. statsd package)"
        raise ConfigurationError(error_msg)

    settings = config.get_settings()
    uri = settings['statsd_url']
    uri = urlparse(uri)

    if settings['project_name'] != '':
        prefix = settings['project_name']
    else:
        prefix = settings['statsd_prefix']

    return Client(uri.hostname, uri.port, prefix)



from kinto import __version__ as VERSION

from kinto.core.testing import unittest

from .support import (BaseWebTest, MINIMALIST_BUCKET,
                      MINIMALIST_GROUP)


class HelloViewTest(BaseWebTest, unittest.TestCase):

    def test_returns_info_about_url_and_version(self):
        response = self.app.get('/')
        self.assertEqual(response.json['project_name'], 'kinto')
        self.assertEqual(response.json['project_version'], VERSION)
        self.assertEqual(response.json['project_docs'],
                         'https://kinto.readthedocs.io/')
        self.assertEqual(response.json['url'], 'http://localhost/v1/')

    def test_hides_user_info_if_anonymous(self):
        response = self.app.get('/')
        self.assertNotIn('user', response.json)

    def test_returns_user_id_if_authenticated(self):
        response = self.app.get('/', headers=self.headers)
        self.assertEqual(response.json['user']['id'], self.principal)

    def test_returns_user_principals_if_authenticated(self):
        group_url = '/buckets/beers/groups/users'
        group = {**MINIMALIST_GROUP}
        group['data']['members'].append(self.principal)
        self.app.put_json('/buckets/beers', MINIMALIST_BUCKET, headers=self.headers)
        self.app.put_json(group_url, group, headers=self.headers)
        response = self.app.get('/', headers=self.headers).json['user']['principals']
        principals = ('system.Everyone', 'system.Authenticated',
                      group_url, self.principal)
        self.assertEqual(sorted(response), sorted(principals))

    def test_capability_is_exposed_if_setting_is_set(self):
        settings = {'experimental_collection_schema_validation': True}
        app = self.make_app(settings=settings)
        resp = app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('schema', capabilities)
        expected = {
            "description": "Validates collection records with JSON schemas.",
            "url": "https://kinto.readthedocs.io/en/latest/api/1.x/"
                   "collections.html#collection-json-schema",
        }
        self.assertEqual(expected, capabilities['schema'])

    def test_capability_is_exposed_if_setting_is_not_set(self):
        settings = self.get_app_settings()
        settings['experimental_collection_schema_validation'] = False
        app = self.make_app(settings=settings)
        resp = app.get('/')
        capabilities = resp.json['capabilities']
        self.assertNotIn('schema', capabilities)

    def test_flush_capability_if_enabled(self):
        settings = {'flush_endpoint_enabled': True}
        app = self.make_app(settings=settings)
        resp = app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('flush_endpoint', capabilities)
        expected = {
            "description": "The __flush__ endpoint can be used to remove "
                           "all data from all backends.",
            "url": "https://kinto.readthedocs.io/en/latest/configuration/"
                   "settings.html#activating-the-flush-endpoint"
        }
        self.assertEqual(expected, capabilities['flush_endpoint'])

    def test_flush_capability_if_not_enabled(self):
        settings = {'flush_endpoint_enabled': False}
        app = self.make_app(settings=settings)
        resp = app.get('/')
        capabilities = resp.json['capabilities']
        self.assertNotIn('flush_endpoint', capabilities)

    def test_permissions_capability_if_enabled(self):
        settings = {'experimental_permissions_endpoint': True}
        app = self.make_app(settings=settings)
        resp = app.get('/')
        capabilities = resp.json['capabilities']
        self.assertIn('permissions_endpoint', capabilities)
        expected = {
            "description": "The permissions endpoint can be used to list "
                           "all user objects permissions.",
            "url": "https://kinto.readthedocs.io/en/latest/configuration/"
                   "settings.html#activating-the-permissions-endpoint"
        }
        self.assertEqual(expected, capabilities['permissions_endpoint'])

    def test_permissions_capability_if_not_enabled(self):
        settings = {'experimental_permissions_endpoint': False}
        app = self.make_app(settings=settings)
        resp = app.get('/')
        capabilities = resp.json['capabilities']
        self.assertNotIn('permissions_endpoint', capabilities)



import mock
from pyramid import httpexceptions

from kinto.core.resource import UserResource, ShareableResource
from kinto.core.storage import exceptions as storage_exceptions

from . import BaseTest


class ResourceTest(BaseTest):

    def test_get_parent_id_default_to_prefixed_userid(self):
        request = self.get_request()
        parent_id = self.resource.get_parent_id(request)
        self.assertEquals(parent_id, 'basicauth:bob')

    def test_raise_if_backend_fails_to_obtain_timestamp(self):
        request = self.get_request()

        with mock.patch.object(request.registry.storage,
                               'collection_timestamp',
                               side_effect=storage_exceptions.BackendError):
            with self.assertRaises(storage_exceptions.BackendError):
                self.resource_class(request)

    def test_raise_unavailable_if_fail_to_obtain_timestamp_with_readonly(self):
        request = self.get_request()

        excepted_exc = httpexceptions.HTTPServiceUnavailable

        request.registry.settings = {'readonly': 'true'}
        with mock.patch.object(request.registry.storage,
                               'collection_timestamp',
                               side_effect=storage_exceptions.BackendError):
            with self.assertRaises(excepted_exc) as cm:
                self.resource_class(request)
                self.assertIn('writable', cm.exception.message)


class ShareableResourceTest(BaseTest):
    resource_class = ShareableResource

    def test_resource_can_be_created_without_context(self):
        try:
            self.resource_class(self.get_request())
        except Exception as e:
            self.fail(e)

    def test_get_parent_id_is_empty(self):
        request = self.get_request()
        parent_id = self.resource.get_parent_id(request)
        self.assertEquals(parent_id, '')


class NewResource(UserResource):
    def get_parent_id(self, request):
        return "overrided"


class ParentIdOverrideResourceTest(BaseTest):
    resource_class = NewResource

    def test_get_parent_can_be_overridded(self):
        request = self.get_request()

        parent_id = self.resource.get_parent_id(request)
        self.assertEquals(parent_id, 'overrided')
        self.assertEquals(self.resource.model.parent_id, 'overrided')



import codecs
import os
from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))


def read_file(filename):
    """Open a related file and return its content."""
    with codecs.open(os.path.join(here, filename), encoding='utf-8') as f:
        content = f.read()
    return content


README = read_file('README.rst')
CHANGELOG = read_file('CHANGELOG.rst')
CONTRIBUTORS = read_file('CONTRIBUTORS.rst')

REQUIREMENTS = [
    'colander >= 1.3.2',
    'colorama',
    'cornice >= 2.4',
    'cornice_swagger >= 0.5',
    'jsonschema',
    'jsonpatch',
    'python-dateutil',
    'pyramid > 1.8',
    'pyramid_multiauth >= 0.8',  # User on policy selected event.
    'transaction',
    'pyramid_tm < 2',
    'requests',
    'structlog >= 16.1.0, < 17',
    'waitress',
    'ujson >= 1.35'
]

POSTGRESQL_REQUIRES = [
    'SQLAlchemy',
    'psycopg2 > 2.5',
    'zope.sqlalchemy',
]

REDIS_REQUIRES = [
    'kinto_redis'
]

SETUP_REQUIRES = [
    'pytest-runner'
]

TEST_REQUIREMENTS = [
    'bravado_core',
    'pytest',
    'WebTest'
]

DEPENDENCY_LINKS = [
]

MONITORING_REQUIRES = [
    'raven',
    'statsd',
    'newrelic',
    'werkzeug',
]

ENTRY_POINTS = {
    'paste.app_factory': [
        'main = kinto:main',
    ],
    'console_scripts': [
        'kinto = kinto.__main__:main'
    ],
}


setup(name='kinto',
      version='6.0.8',
      description='Kinto Web Service - Store, Sync, Share, and Self-Host.',
      long_description="{}\n\n{}\n\n{}".format(README, CHANGELOG, CONTRIBUTORS),
      license='Apache License (2.0)',
      classifiers=[
          "Programming Language :: Python",
          "Programming Language :: Python :: 3",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6",
          "Programming Language :: Python :: Implementation :: CPython",
          "Topic :: Internet :: WWW/HTTP",
          "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
          "License :: OSI Approved :: Apache Software License"
      ],
      keywords="web sync json storage services",
      author='Mozilla Services',
      author_email='storage-team@mozilla.com',
      url='https://github.com/Kinto/kinto',
      packages=find_packages(),
      package_data={'': ['*.rst', '*.py', '*.yaml']},
      include_package_data=True,
      zip_safe=False,
      setup_requires=SETUP_REQUIRES,
      tests_require=TEST_REQUIREMENTS,
      install_requires=REQUIREMENTS,
      extras_require={
          'redis': REDIS_REQUIRES,
          'postgresql': POSTGRESQL_REQUIRES,
          'monitoring': MONITORING_REQUIRES,
      },
      test_suite="tests",
      dependency_links=DEPENDENCY_LINKS,
      entry_points=ENTRY_POINTS)



from bravado_core.response import validate_response

from .support import SwaggerTest, MINIMALIST_RECORD


class SwaggerRecordResponsesTest(SwaggerTest):

    def test_get_record_200(self):
        response = self.app.get('/buckets/b1/collections/c1/records/r1',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].get_record
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_record_200(self):
        response = self.app.post_json('/buckets/b1/collections/c1/records',
                                      self.record, headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].create_record
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_post_record_201(self):
        response = self.app.post_json('/buckets/b1/collections/c1/records',
                                      MINIMALIST_RECORD, headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].create_record
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_put_record_200(self):
        response = self.app.put_json('/buckets/b1/collections/c1/records/r1',
                                     MINIMALIST_RECORD, headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].update_record
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_put_record_201(self):
        response = self.app.put_json('/buckets/b1/collections/c1/records/r2',
                                     MINIMALIST_RECORD, headers=self.headers, status=201)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].update_record
        schema = self.spec.deref(op.op_spec['responses']['201'])
        validate_response(schema, op, response)

    def test_delete_record_200(self):
        response = self.app.delete('/buckets/b1/collections/c1/records/r1',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].delete_record
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_get_records_200(self):
        response = self.app.get('/buckets/b1/collections/c1/records',
                                headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].get_records
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)

    def test_delete_records_200(self):
        response = self.app.delete('/buckets/b1/collections/c1/records',
                                   headers=self.headers, status=200)
        response = self.cast_bravado_response(response)
        op = self.resources['Records'].delete_records
        schema = self.spec.deref(op.op_spec['responses']['200'])
        validate_response(schema, op, response)
